
from typing import Tuple, List

import pandas as pd
import datetime
from sqlalchemy import func
import statsmodels.api as sm
from statsmodels import regression
import numpy as np
from ...util.singleton import Singleton
from ..wrapper.mysql import DerivedDatabaseConnector
from .raw import RawDataApi
from .basic import BasicDataApi
from ..view.derived_models import *

class DerivedDataApi(metaclass=Singleton):
    def get_fund_indicator(self, fund_list):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                        FundIndicator.fund_id,
                        FundIndicator.datetime,
                        FundIndicator.alpha,
                        FundIndicator.beta,
                        FundIndicator.fee_rate,
                        FundIndicator.track_err,
                    ).filter(
                        FundIndicator.fund_id.in_(fund_list),
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundIndicator.__tablename__}')

    def delete_fund_indicator_with_dt(self, date_list):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FundIndicator
                ).filter(
                    FundIndicator.datetime.in_(date_list),
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FundIndicator.__tablename__}')
                return False

    def get_fund_indicator_by_date(self, fund_list, date):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                    FundIndicator
                ).filter(
                    FundIndicator.datetime == date,
                    FundIndicator.fund_id.in_(fund_list),
                )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df

            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundIndicator.__tablename__}')

    def get_fund_indicator_monthly(self, start_date, end_date, fund_list, columns: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                    FundIndicatorMonthly.fund_id,
                    FundIndicatorMonthly.datetime,
                )
                if columns:
                    query = query.add_columns(*columns)
                query = query.filter(
                    FundIndicatorMonthly.fund_id.in_(fund_list),
                    FundIndicatorMonthly.datetime >= start_date,
                    FundIndicatorMonthly.datetime <= end_date
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundIndicatorMonthly.__tablename__}')

    def get_fund_indicator_group(self, start_date: str, end_date: str, data_cycle: str = ''):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                    FundIndicatorGroup
                )
                query = query.filter(
                    FundIndicatorGroup.datetime >= start_date,
                    FundIndicatorGroup.datetime <= end_date
                )
                if data_cycle:
                    query = query.filter(
                        FundIndicatorGroup.data_cycle == data_cycle
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundIndicatorGroup.__tablename__}')

    def get_latest_fund_indicator_group(self, data_cycle: str = ''):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                sub_q = quant_session.query(
                    FundIndicatorGroup.datetime.label('latest_time'),
                )
                if data_cycle:
                    sub_q = sub_q.filter(
                        FundIndicatorGroup.data_cycle == data_cycle,
                    )
                sub_q = sub_q.order_by(
                    FundIndicatorGroup.datetime.desc(),
                ).limit(1).subquery()

                query = quant_session.query(FundIndicatorGroup).filter(
                    FundIndicatorGroup.datetime == sub_q.c.latest_time,
                )
                if data_cycle:
                    query = query.filter(
                        FundIndicatorGroup.data_cycle == data_cycle,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundIndicatorGroup.__tablename__}')

    def get_fund_score_extended_by_id(self, fund_id, data_cycle: str = '1Y'):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                    FundScoreExtended,
                ).filter(
                    FundScoreExtended.fund_id == fund_id,
                    FundScoreExtended.data_cycle == data_cycle,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundScoreExtended.__tablename__}')
                return pd.DataFrame([])

    def get_latest_fund_score_new_by_id(self, fund_id, data_cycle: str = '1Y'):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                    FundScoreNew,
                ).filter(
                    FundScoreNew.fund_id == fund_id,
                    FundScoreNew.data_cycle == data_cycle,
                ).order_by(
                    FundScoreNew.datetime.desc()
                ).limit(1)

                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundScoreNew.__tablename__}')
                return pd.DataFrame([])

    def get_fund_score_extended(self, start_date: str, end_date: str, fund_list: Tuple[str] = (), columns: Tuple[str] = (), data_cycle: str = '1Y'):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                if columns:
                    query = quant_session.query(
                        FundScoreExtended.fund_id,
                        FundScoreExtended.datetime,
                    )
                    query = query.add_columns(*columns)
                else:
                    query = quant_session.query(
                        FundScoreExtended,
                    )
                if fund_list:
                    query = query.filter(
                        FundScoreExtended.fund_id.in_(fund_list)
                    )
                query = query.filter(
                    FundScoreExtended.datetime >= start_date,
                    FundScoreExtended.datetime <= end_date,
                    FundScoreExtended.data_cycle == data_cycle,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundScoreExtended.__tablename__}')

    def get_fund_score_extended_for_ranks(self, start_date: str, end_date: str, fund_list: Tuple[str] = ()):
        return self.get_fund_score_extended(start_date, end_date, fund_list=fund_list, columns=['wind_class_1', 'return_score', 'robust_score', 'timing_score', 'return_rank', 'robust_rank', 'timing_rank'])

    def get_index_valuation_develop(self, index_ids, start_date, end_date):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexValuationLongTerm
                ).filter(
                    IndexValuationLongTerm.index_id.in_(index_ids),
                    IndexValuationLongTerm.datetime >= start_date,
                    IndexValuationLongTerm.datetime <= end_date,
                ).order_by(IndexValuationLongTerm.datetime.asc())
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexValuationLongTerm.__tablename__}')
                return pd.DataFrame([])

    def get_index_pair(self):
        index_val = {'PE':'pe_ttm','PB':'pb_mrq','ROE':'roe'}
        return index_val

    def get_index_val_info(self):
        index_info = BasicDataApi().get_index_info()
        desc_df = index_info.set_index('index_id').to_dict()['desc_name']
        info_dic = {
            'A股指数':['sse50','hs300','csi500','gem_50'],
            '港股指数':['hsi','hsi_stateown','hsi_ita','hsi_china25','hsi_mssize','hsi_shkscne','hsi_intertec'],
            '风格指数':['csi300_growth','csi300_value','csi_grow500','csi_v500','csi800_value','csi800_growth'],
        }
        info_list = ['行业指数','主题指数']
        asset_info = BasicDataApi().get_asset_info_by_type(info_list)
        result = []
        for asset_id, index_list in info_dic.items():
            dic = {'title':asset_id,'items':[]}
            for index_id in index_list:
                _dic = {
                    'id':index_id,
                    'name':desc_df[index_id],
                }
                dic['items'].append(_dic)
            result.append(dic)

        for asset_id in info_list:
            _df = asset_info[asset_info.asset_type == asset_id]
            dic = {'title':asset_id,'items':[]}
            for i in _df.itertuples():
                _dic = {
                    'id':i.real_id,
                    'name':i.asset_name,
                }
                dic['items'].append(_dic)
            result.append(dic)
        index_val = self.get_index_pair()
        data = {'index_group':result,'index_val':list(index_val.keys())}
        return data

    def get_index_valuation_with_period(self, begin_date, end_date, time_para, index_list, valuation_type='PE'):
        index_val = valuation_type
        data_type = f'{index_val}估值'
        inputs_dic = self.get_index_pair()
        columns = [inputs_dic[index_val]]
        
        begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexValuationLongTerm.index_id,
                    IndexValuationLongTerm.datetime,
                ).add_columns(*columns).filter(
                    IndexValuationLongTerm.index_id.in_(index_list),
                    IndexValuationLongTerm.datetime >= begin_date,
                    IndexValuationLongTerm.datetime <= end_date,
                ).order_by(IndexValuationLongTerm.datetime.asc())
                df = pd.read_sql(query.statement, query.session.bind)
                l = df.columns.tolist()
                value_col = [i for i in l if i not in ['datetime','index_id']][0]
                index_info = BasicDataApi().get_index_info(index_list)
                desc_dic = index_info.set_index('index_id').to_dict()['desc_name']
                df = df.pivot_table(columns='index_id',values=value_col,index='datetime')
                _df = df.copy()
                df.loc['中位数',:] = _df.median()
                df.loc['最大值',:] = _df.max()
                df.loc['最小值',:] = _df.min()
                df.loc['现值',:] = _df.ffill().iloc[-1]
                df = df.rename(columns=desc_dic)
                df = df.loc[['中位数','最大值','最小值','现值']].round(2)   
                cols = df.columns.tolist()
                name_dic = {i:i.replace('(申万)','') for i in cols}
                df = df.rename(columns=name_dic).rename(columns={'HKCTMT人民币':'沪港通TMT'}) 
                return {'data':df,'title':data_type}
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexValuationLongTerm.__tablename__}')
                return pd.DataFrame([])

    def get_index_valuation_last_date(self, dt, index_ids=[]):
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                dt = db_session.query(
                    func.max(IndexValuationLongTerm.datetime)).filter(
                            IndexValuationLongTerm.index_id.in_(index_ids), 
                            IndexValuationLongTerm.datetime <= dt,
                ).one_or_none()[0]
                query = db_session.query(IndexValuationLongTerm).filter(
                        IndexValuationLongTerm.datetime == dt,
                        IndexValuationLongTerm.index_id.in_(index_ids),
                    )
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexValuationLongTerm.__tablename__}')
                return pd.DataFrame([])

    def get_index_valuation_develop_without_date(self, index_ids):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexValuationLongTerm
                ).filter(
                    IndexValuationLongTerm.index_id.in_(index_ids),
                ).order_by(IndexValuationLongTerm.datetime.asc())
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexValuationLongTerm.__tablename__}')
                return pd.DataFrame([])

    def get_index_valuation_develop_columns_by_id(self, index_id, columns):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    *columns
                ).filter(
                    IndexValuationLongTerm.index_id == index_id,
                ).order_by(IndexValuationLongTerm.datetime.asc())
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {IndexValuationLongTerm.__tablename__}')
                return pd.DataFrame([])

    def delete_index_valuation(self, date_to_delete: datetime.date, index_id_list: List[str]) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    IndexValuationLongTerm
                ).filter(
                    IndexValuationLongTerm.index_id.in_(index_id_list),
                    IndexValuationLongTerm.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {IndexValuationLongTerm.__tablename__}')
                return False

    def get_asset_allocation_info(self, version:int=2):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    AssetAllocationInfo
                ).filter(
                    AssetAllocationInfo.version == version
                ).order_by(AssetAllocationInfo.allocation_id)
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {AssetAllocationInfo.__tablename__}')
                return pd.DataFrame([])

    def get_style_factor_return(self, start_date: str, end_date: str, index_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    StyleAnalysisFactorReturn
                )
                if index_list:
                    query = query.filter(StyleAnalysisFactorReturn.universe_index.in_(index_list))
                query = query.filter(
                    StyleAnalysisFactorReturn.datetime >= start_date,
                    StyleAnalysisFactorReturn.datetime <= end_date,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {StyleAnalysisFactorReturn.__tablename__}')
                return

    def get_barra_cne5_factor_return(self, start_date: str, end_date: str):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    BarraCNE5FactorReturn
                ).filter(
                    BarraCNE5FactorReturn.datetime >= start_date,
                    BarraCNE5FactorReturn.datetime <= end_date,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {BarraCNE5FactorReturn.__tablename__}')
                return

    def get_allocation_distribution(self, version:str=1):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    AllocationDistribution
                ).filter(
                    AllocationDistribution.version == version
                )
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {AllocationDistribution.__tablename__}')
                return

    def get_fund_manager_index(self, manager_id: Tuple[str] = (), start_date: str = '', end_date: str = '', fund_type: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    FundManagerIndex
                )
                if manager_id:
                    query = query.filter(FundManagerIndex.manager_id.in_(manager_id))
                if start_date:
                    query = query.filter(FundManagerIndex.datetime >= start_date)
                if end_date:
                    query = query.filter(FundManagerIndex.datetime <= end_date)
                if fund_type:
                    query = query.filter(FundManagerIndex.fund_type.in_(fund_type))
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundManagerIndex.__tablename__}')

    def get_fund_manager_score(self, manager_id: Tuple[str] = (), start_date: str = '', end_date: str = '', fund_type: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    FundManagerScore
                )
                if manager_id:
                    query = query.filter(FundManagerScore.manager_id.in_(manager_id))
                if start_date:
                    query = query.filter(FundManagerScore.datetime >= start_date)
                if end_date:
                    query = query.filter(FundManagerScore.datetime <= end_date)
                if fund_type:
                    query = query.filter(FundManagerScore.fund_type.in_(fund_type))
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundManagerScore.__tablename__}')

    def get_latest_fund_style_box(self):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                sub_q = mn_session.query(
                    StyleBox.fund_id.label('temp_id'),
                    func.max(StyleBox.datetime).label('temp_date'),
                ).group_by(StyleBox.fund_id).subquery()

                query = mn_session.query(
                    StyleBox.fund_id,
                    StyleBox.x,
                    StyleBox.y,
                    StyleBox.datetime,
                ).filter(
                    StyleBox.fund_id == sub_q.c.temp_id,
                    StyleBox.datetime == sub_q.c.temp_date
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_latest_fund_style_box {e} from {StyleBox.__tablename__}')
                return

    def get_funds_latest_fund_style_box(self, fund_list:Tuple[str]):
        # x： 0 价值型 40 平衡型 60 成长型 100
        # y： 0 小盘 25 中盘 50 大盘 100
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                sub_q = mn_session.query(
                    StyleBox.fund_id.label('temp_id'),
                    func.max(StyleBox.datetime).label('temp_date'),
                ).filter(
                    StyleBox.fund_id.in_(fund_list),
                ).group_by(StyleBox.fund_id).subquery()

                query = mn_session.query(
                    StyleBox.fund_id,
                    StyleBox.x,
                    StyleBox.y,
                    StyleBox.datetime,
                ).filter(
                    StyleBox.fund_id == sub_q.c.temp_id,
                    StyleBox.datetime == sub_q.c.temp_date
                )
                df = pd.read_sql(query.statement, query.session.bind)
                df['x'] = df['x'].map(lambda x: 0 if x < 0 else x)
                df['x'] = df['x'].map(lambda x: 300 if x > 300 else x)
                df['y'] = df['y'].map(lambda y: 0 if y < 0 else y)
                df['y'] = df['y'].map(lambda y: 400 if y > 400 else y)
                df['x'] = df['x'] / 3
                df['y'] = df['y'] / 4
                return df
            except Exception as e:
                print(f'Failed to get_latest_fund_style_box {e} from {StyleBox.__tablename__}')
                return

    def get_fund_style_box(self, fund_id: Tuple[str] = (), datetime: str = ''):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    StyleBox
                )
                if fund_id:
                    query = query.filter(
                        StyleBox.fund_id.in_(fund_id),
                    )
                if datetime:
                    query = query.filter(
                        StyleBox.datetime == datetime,
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_fund_style_box {e} from {StyleBox.__tablename__}')
                return

    def get_new_share_fund_rank(self):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    NewShareFundRank
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_new_share_fund_rank {e} from {NewShareFundRank.__tablename__}')
                return

    def get_convertible_bond_fund_rank(self):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    ConvertibleBondFundRank
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_convertible_bond_fund_rank {e} from {ConvertibleBondFundRank.__tablename__}')
                return

    def get_abs_return_fund_rank(self):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    AbsReturnFundRank
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_abs_return_fund_rank {e} from {AbsReturnFundRank.__tablename__}')
                return

    def delete_fund_style_box(self, date_to_delete: datetime.date, fund_list: List[str]):
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                db_session.query(
                    StyleBox
                ).filter(
                    StyleBox.fund_id.in_(fund_list),
                    StyleBox.datetime == date_to_delete,
                ).delete(synchronize_session=False)
                db_session.commit()
                return True
            except Exception as e:
                print(f'Failed to delete data <err_msg> {e} from {StyleBox.__tablename__}')

    def get_fund_manager_info(self, mng_list: Tuple[str] = (), fund_list: Tuple[str] = (), columns: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                if columns:
                    query = quant_session.query(
                        FundManagerInfo.mng_id,
                        FundManagerInfo.start_date,
                        FundManagerInfo.fund_id,
                    ).add_columns(*columns)
                else:
                    query = quant_session.query(
                        FundManagerInfo,
                    )
                if mng_list:
                    query = query.filter(
                        FundManagerInfo.mng_id.in_(mng_list),
                    )
                if fund_list:
                    query = query.filter(
                        FundManagerInfo.fund_id.in_(fund_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get data <err_msg> {e} from {FundManagerInfo.__tablename__}')

    def get_market_portfolio_indicator(self):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    ThirdPartyPortfolioIndicator
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_market_portfolio_indicator {e} from {ThirdPartyPortfolioIndicator.__tablename__}')
                return pd.DataFrame([])

    def get_market_portfolio_info(self):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    ThirdPartyPortfolioInfo
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_market_portfolio_info {e} from {ThirdPartyPortfolioInfo.__tablename__}')
                return pd.DataFrame([])

    def get_market_portfolio_trade_dates(self, po_id):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    ThirdPartyPortfolioTrade.datetime
                ).filter(
                    ThirdPartyPortfolioTrade.po_id == po_id,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_market_portfolio_trade_dates {e} from {ThirdPartyPortfolioTrade.__tablename__}')
                return pd.DataFrame([])

    def get_market_portfolio_trade_by_date(self, po_id, date):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    ThirdPartyPortfolioTrade
                ).filter(
                    ThirdPartyPortfolioTrade.po_id == po_id,
                    ThirdPartyPortfolioTrade.datetime == date,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_market_portfolio_trade_by_date {e} from {ThirdPartyPortfolioTrade.__tablename__}')
                return pd.DataFrame([])

    def get_market_portfolio_position(self, po_id):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    ThirdPartyPortfolioPositionLatest
                ).filter(
                    ThirdPartyPortfolioPositionLatest.po_id == po_id,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_market_portfolio_position {e} from {ThirdPartyPortfolioPositionLatest.__tablename__}')
                return pd.DataFrame([])

    def get_market_portfolio_position_by_src(self, po_srcs: Tuple[int] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    ThirdPartyPortfolioPositionLatest
                )
                if po_srcs:
                    query = query.filter(
                        ThirdPartyPortfolioPositionLatest.po_src.in_(po_srcs),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_market_portfolio_position_by_src {e} from {ThirdPartyPortfolioPositionLatest.__tablename__}')
                return

    def get_stock_factor_info(self):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    StockFactorInfo
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_stock_factor_info {e} from {StockFactorInfo.__tablename__}')
                return

    def get_fund_industry_exposure(self, fund_id):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                sub_q = session.query(
                    FundIndicatorGroup.datetime.label('latest_time'),
                ).order_by(
                    FundIndicatorGroup.datetime.desc(),
                ).limit(1).subquery()

                query = session.query(
                    FundIndustryExposure
                ).filter(
                    FundIndustryExposure.fund_id == fund_id,
                    FundIndustryExposure.datetime == sub_q.c.latest_time,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'Failed to get_fund_industry_exposure {e} from {FundIndicatorGroup.__tablename__}')
                return pd.DataFrame([])

    def get_fof_nav(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    FOFNav,
                ).filter(
                    FOFNav.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFNav.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_nav <err_msg> {e} from {FOFNav.__tablename__}')

    def delete_fof_nav(self, date_to_delete: datetime.date, manager_id: str, fof_id_list: List[str]) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFNav
                ).filter(
                    FOFNav.fof_id.in_(fof_id_list),
                    FOFNav.datetime == date_to_delete,
                    FOFNav.manager_id == manager_id,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFNav.__tablename__}')
                return False

    def get_fof_nav_public(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    FOFNavPublic,
                ).filter(
                    FOFNavPublic.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFNavPublic.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_nav_public <err_msg> {e} from {FOFNavPublic.__tablename__}')

    def get_fof_nav_calc(self, manager_id, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    FOFNavCalc,
                ).filter(
                    FOFNavCalc.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFNavCalc.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_nav_calc <err_msg> {e} from {FOFNavCalc.__tablename__}')

    def delete_fof_nav_calc(self, manager_id: str, fof_id: str, date_list: List[datetime.date]) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFNavCalc
                ).filter(
                    FOFNavCalc.manager_id == manager_id,
                    FOFNavCalc.fof_id == fof_id,
                    FOFNavCalc.datetime.in_(date_list),
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFNavCalc.__tablename__}')
                return False

    def get_fof_position(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                query = session.query(
                    FOFPosition,
                ).filter(
                    FOFPosition.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFPosition.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_position <err_msg> {e} from {FOFPosition.__tablename__}')

    def delete_fof_position(self, manager_id: str, fof_id: str, date_list: List[datetime.date]) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFPosition
                ).filter(
                    FOFPosition.manager_id == manager_id,
                    FOFPosition.fof_id == fof_id,
                    FOFPosition.datetime.in_(date_list),
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFPosition.__tablename__}')
                return False

    def get_fof_investor_data(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFInvestorData,
                ).filter(
                    FOFInvestorData.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFInvestorData.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_investor_data <err_msg> {e} from {FOFInvestorData.__tablename__}')

    def delete_fof_investor_data(self, manager_id: str, fof_id_to_delete: str, investor_id_list: List[str]) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFInvestorData
                ).filter(
                    FOFInvestorData.fof_id == fof_id_to_delete,
                    FOFInvestorData.manager_id == manager_id,
                    FOFInvestorData.investor_id.in_(investor_id_list),
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFInvestorData.__tablename__}')
                return False

    def get_fof_position_detail(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    FOFPositionDetail,
                ).filter(
                    FOFPositionDetail.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        FOFPositionDetail.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_fof_position_detail <err_msg> {e} from {FOFPositionDetail.__tablename__}')

    def delete_fof_position_detail(self, manager_id: str, fof_id_to_delete: str, fund_id_list: List[str]) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    FOFPositionDetail
                ).filter(
                    FOFPositionDetail.fof_id == fof_id_to_delete,
                    FOFPositionDetail.manager_id == manager_id,
                    FOFPositionDetail.fund_id.in_(fund_id_list),
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {FOFPositionDetail.__tablename__}')
                return False

    def get_hedge_fund_investor_pur_redemp(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    HedgeFundInvestorPurAndRedemp,
                ).filter(
                    HedgeFundInvestorPurAndRedemp.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        HedgeFundInvestorPurAndRedemp.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_hedge_fund_investor_pur_redemp <err_msg> {e} from {HedgeFundInvestorPurAndRedemp.__tablename__}')

    def get_hedge_fund_investor_pur_redemp_by_id(self, id: int):
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    HedgeFundInvestorPurAndRedemp,
                ).filter(
                    HedgeFundInvestorPurAndRedemp.id == id,
                )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_hedge_fund_investor_pur_redemp_by_id <err_msg> {e} from {HedgeFundInvestorPurAndRedemp.__tablename__}')

    def delete_hedge_fund_investor_pur_redemp(self, id_to_delete: int) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    HedgeFundInvestorPurAndRedemp
                ).filter(
                    HedgeFundInvestorPurAndRedemp.id == id_to_delete,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {HedgeFundInvestorPurAndRedemp.__tablename__}')
                return False

    def get_hedge_fund_investor_div_carry(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    HedgeFundInvestorDivAndCarry,
                ).filter(
                    HedgeFundInvestorDivAndCarry.manager_id == manager_id,
                )
                if fof_id_list:
                    query = query.filter(
                        HedgeFundInvestorDivAndCarry.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_hedge_fund_investor_div_carry <err_msg> {e} from {HedgeFundInvestorDivAndCarry.__tablename__}')

    def delete_hedge_fund_investor_div_carry(self, id_to_delete: int) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    HedgeFundInvestorDivAndCarry
                ).filter(
                    HedgeFundInvestorDivAndCarry.id == id_to_delete,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {HedgeFundInvestorDivAndCarry.__tablename__}')
                return False

    def get_hedge_fund_custodian_data(self, manager_id: str, fof_id_list: Tuple[str] = ()):
        with DerivedDatabaseConnector().managed_session() as db_session:
            try:
                query = db_session.query(
                    HedgeFundCustodianData,
                )
                if fof_id_list:
                    query = query.filter(
                        HedgeFundCustodianData.fof_id.in_(fof_id_list),
                    )
                return pd.read_sql(query.statement, query.session.bind)
            except Exception as e:
                print(f'failed to get_hedge_fund_custodian_data <err_msg> {e} from {HedgeFundCustodianData.__tablename__}')

    def delete_hedge_fund_custodian_data(self, manager_id: str, fof_id: str, datetime) -> bool:
        with DerivedDatabaseConnector().managed_session() as session:
            try:
                session.query(
                    HedgeFundCustodianData
                ).filter(
                    HedgeFundCustodianData.manager_id == manager_id,
                    HedgeFundCustodianData.fof_id == fof_id,
                    HedgeFundCustodianData.datetime == datetime,
                ).delete(synchronize_session=False)
                session.commit()
                return True
            except Exception as e:
                print(f'failed to delete data <err_msg> {e} from {HedgeFundCustodianData.__tablename__}')
                return False

    def get_style_factor_ret(self, begin_date, end_date, time_para, index_list=['hs300']):
        try:
            start_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            df = self.get_style_factor_return(start_date, end_date, [index_list])
            dic = {
                'latest_size':'规模',
                'bp':'价值',
                'short_term_momentum':'短期动量',
                'long_term_momentum':'长期动量',
                'high_low':'波动率',
            }
            df = df.set_index('datetime')[['latest_size','bp','short_term_momentum','long_term_momentum','high_low','const']]
            df = df[list(dic.keys())].rename(columns=dic)
            df = (df+1).cumprod()
            df = df/df.iloc[0] - 1
            df.index.name = 'datetime'
            return df
        except Exception as e:
            print(f'failed to get data <err_msg> {e} from derived.get_style_factor_ret')
            return False

    def get_industry_stock_ret_and_size(self, begin_date, end_date, time_para):
        try:
            def process_x(x):
                x.loc[:,'weight'] = x.total_share / x.total_share.sum()
                x = x[x.weight >= 0.015]
                return x
            # 建议只用短时间，上时间慢
            raw_api = RawDataApi()
            basic_api = BasicDataApi()
            begin_date, end_date = raw_api.get_date_range(time_para, begin_date, end_date)
            stock_df = raw_api.get_em_stock_info()
            today = datetime.date.today()
            _b_d = today - datetime.timedelta(days=90)
            stock_df = stock_df[stock_df.list_date < _b_d]
            stock_df = stock_df[['stock_id','name','bl_sws_ind_code']].dropna()
            stock_df.loc[:,'industry_1'] = stock_df.bl_sws_ind_code.map(lambda x: x.split('-')[0])
            stock_df.loc[:,'industry_2'] = stock_df.bl_sws_ind_code.map(lambda x: x.split('-')[1])
            stock_df = stock_df.drop(columns=['bl_sws_ind_code']).set_index('stock_id')
            industry_df = raw_api.get_em_industry_info()
            industry_dic_name = industry_df.set_index('em_id').to_dict()['ind_name']
            stock_df.loc[:,'industry_1_name'] = stock_df.industry_1.map(industry_dic_name)
            stock_df.loc[:,'industry_2_name'] = stock_df.industry_2.map(industry_dic_name)
            stock_list = stock_df.index.tolist()
            total_share = raw_api.get_em_daily_info_last_date(stock_list=stock_list,columns=['total_share'])
            total_share = total_share.pivot_table(index='stock_id',columns='datetime',values='total_share')
            total_share.columns = ['total_share']
            groupby_df = stock_df.join(total_share).copy()
            stock_df = groupby_df.groupby('industry_2').apply(lambda x: process_x(x))
            stock_df = stock_df.drop(columns=['industry_2']).reset_index().set_index('stock_id')
            stock_list = stock_df.index.tolist()
            stock_price = raw_api.get_em_stock_price(start_date=begin_date, end_date=end_date,stock_list=stock_list,columns=['close'])
            last_date = sorted(stock_price.datetime.unique().tolist())[-1]
            last_price = stock_price[stock_price['datetime'] == last_date]
            size_df = last_price.set_index('stock_id')[['close']].join(total_share).dropna()
            size_df.loc[:,'size'] = size_df.close*size_df.total_share
            stock_df = stock_df.join(size_df[['size']]).dropna()
            indus_size_1 = stock_df[['industry_1','size']].groupby('industry_1').sum()
            indus_size_2 = stock_df[['industry_2','size']].groupby('industry_2').sum()
            stock_price = stock_price.pivot_table(index='datetime',columns='stock_id',values='close').ffill()
            stock_ret = pd.DataFrame(stock_price.iloc[-1] / stock_price.iloc[0] - 1)
            stock_ret.columns=['stock_ret']
            stock_df = stock_df.join(stock_ret).dropna()
            industry_list = indus_size_1.index.tolist() + indus_size_2.index.tolist()
            index_info = basic_api.get_index_info_by_em_id(em_id_list=industry_list)
            industry_dic = index_info.set_index('index_id').to_dict()['em_id']
            index_list = list(industry_dic.keys())
            industry_price = basic_api.get_index_price_dt(start_date=begin_date,end_date=end_date,index_list=index_list,columns=['close'])
            industry_price = industry_price.pivot_table(columns='index_id',values='close',index='datetime').ffill()
            industry_price = industry_price.rename(columns=industry_dic)
            indus_ret = pd.DataFrame(industry_price.iloc[-1] / industry_price.iloc[0] - 1)
            indus_ret.columns= ['industry_ret']
            indus_size_1 = indus_size_1.join(indus_ret).dropna()
            indus_size_2 = indus_size_2.join(indus_ret).dropna()
            indus_size_1.loc[:,'industry_1_name'] = indus_size_1.index.map(industry_dic_name)
            indus_size_2.loc[:,'industry_2_name'] = indus_size_2.index.map(industry_dic_name)
            _df = stock_df.reset_index().set_index('industry_2')[['industry_1']]
            _df = _df[~_df.index.duplicated(keep='first')]
            indus_size_2 = indus_size_2.join(_df)
            data = {
                '股票收益':stock_df,
                '行业一级收益':indus_size_1,
                '行业二级收益':indus_size_2,
            }
            return data
        except Exception as e:
            print(f'failed to get data <err_msg> {e} from derived.get_industry_stock_ret_and_size')
            return False
        
    def get_all_industry_val(self, begin_date, end_date, time_para, value_type='估值',industry_type='一级行业'):
        try:
            industry_type_dic = {
                '一级行业':1,
                '二级行业':2
            }
            dic = {
                'pb_mrq':'PB',
                'roe':'ROE'
            }
            raw_api = RawDataApi()
            begin_date, end_date = raw_api.get_date_range(time_para, begin_date, end_date)
            industry_info_raw = RawDataApi().get_em_industry_info(ind_class_type=industry_type_dic[industry_type])
            em_id_raw = industry_info_raw.em_id.tolist()
            index_info = BasicDataApi().get_index_info_by_em_id(em_id_list=em_id_raw)
            index_info.desc_name = index_info.desc_name.map(lambda x: x.replace('(申万)',''))
            index_ids = index_info.index_id.tolist()
            asset_name = index_info.set_index('index_id').to_dict()['desc_name']
            if value_type == '估值':
                df = self.get_index_valuation_last_date(dt=end_date, index_ids=index_ids)
                df = df.set_index('index_id')[dic.keys()].rename(columns=dic).round(2)
                df.loc[:,'desc_name'] = df.index.map(asset_name)
                df = df.reset_index(drop=True)
            elif value_type == '估值分位':
                df = DerivedDataApi().get_index_valuation_develop(end_date=end_date,start_date=begin_date,index_ids=index_ids)
                index_pb = df.pivot_table(index='datetime',columns='index_id',values='pb_mrq')
                index_roe = df.pivot_table(index='datetime',columns='index_id',values='roe')
                pb_rank = index_pb.rank(pct=True, axis=0).iloc[-1]
                pb_rank.name = 'PB'
                roe_rank = index_roe.rank(pct=True, axis=0).iloc[-1]
                roe_rank.name = 'ROE'
                df = pd.concat([pb_rank, roe_rank],axis=1)
                df.loc[:,'desc_name'] = df.index.map(asset_name)
            
            df = df.replace({np.Inf:np.nan,-np.Inf:np.nan}).dropna()
            _x = df.ROE.values
            x = sm.add_constant(_x)
            y = df.PB.values
            model = regression.linear_model.OLS(y,x).fit()
            c = model.params[0]
            k = model.params[1]
            x_array = np.arange(min(_x),max(_x), (max(_x)-min(_x))/100)
            y_array = k*x_array + c
            df_line = pd.DataFrame({'ROE':x_array,'PB':y_array,}).round(2)
            df['PB'] = df['PB'].round(2)
            df['ROE'] = df['ROE'].round(2)
            data = {
                '散点':df,
                '虚线':df_line,
            }
            return data
        except Exception as e:
            print(f'failed to get data <err_msg> {e} from derived.get_all_industry_val')
            return False

    def index_val_history_info(self):   
        try:
            asset_list = ['行业指数','主题指数']
            index_info = BasicDataApi().get_asset_info_by_type(asset_list)
            result = {'大类资产':[['hs300','沪深300'],['csi500','中证500'],['gem','创业板']]}
            for asset_id in asset_list:
                _dfi = index_info[index_info['asset_type'] == asset_id]
                result[asset_id]=[]
                for r in _dfi.itertuples():
                    result[asset_id].append([r.real_id,r.asset_name])
            return result
        except Exception as e:
            print(f'failed to get data <err_msg> {e} from derived.index_val_history_info')
            return False

    def stock_debt_val_info(self):
        index_list = ['sse50','hs300','csi500','csi1000','gem']
        index_info = BasicDataApi().get_index_info(index_list)
        index_info = [{'id':r.index_id, 'name':r.desc_name} for r in index_info.itertuples()]
        valuation_type = [{'id':'PE','name':'风险溢价'},{'id':'DY','name':'股息率溢价'}]
        debt_ret_list = [   {'id':'TB_1Y','name':'1年国债收益率'},
                            {'id':'TB_3Y','name':'3年国债收益率'},
                            {'id':'TB_5Y','name':'5年国债收益率'},
                            {'id':'TB_10Y','name':'10年国债收益率'}]
        data = {
            '股指':index_info,
            '溢价方式':valuation_type,
            '债指':debt_ret_list,
        }
        return data

    def stock_debt_val_detail(self, index_id, valuation_type,debt_ret_index):
        try:
            dic = {
                'pe_ttm':'PE',
                'dy':'DY',
            }
            title_dic = {
                'PE':'风险溢价',
                'DY':'股息率溢价',
            }
            bond_yield = RawDataApi().get_em_macroeconomic_daily(codes=[debt_ret_index])
            index_risk = DerivedDataApi().get_index_valuation_develop_without_date(index_ids=[index_id])
            index_info = BasicDataApi().get_index_info([index_id])
            index_name = index_info.set_index('index_id').loc[index_id,'desc_name']

            index_risk = index_risk.set_index('datetime').rename(columns=dic)[[valuation_type]].join(bond_yield.set_index('datetime').rename(columns={'value':'debt_ret'})['debt_ret']).ffill()

            pctrank = lambda x: x.rank(pct=True).iloc[-1]
            if valuation_type == 'PE':
                index_risk.loc[:,'reverse_val'] = 1 / index_risk[valuation_type]
            else:
                index_risk.loc[:,'reverse_val'] = index_risk[valuation_type]
            index_risk.loc[:,'debt_ret'] = index_risk.debt_ret / 100
            index_risk.loc[:,'risk_value'] = index_risk.reverse_val - index_risk.debt_ret

            min_periods = int(3*242)
            window = index_risk.shape[0]
            window= 5 * 242
            index_risk['risk_values_pct'] = index_risk.risk_value.rolling(window=window, min_periods=min_periods).apply(pctrank, raw=False)
            index_col_pct = 'risk_values_pct'
            index_col = 'risk_value'
            diff = 0.00
            a1 = index_risk[index_risk[index_col_pct] >= (0.90 - diff)][[index_col]].rename(columns={index_col:'区间分位最高10%'})
            a2 = index_risk[(index_risk[index_col_pct] <= (0.90 + diff))&(index_risk[index_col_pct] >= 0.7-diff)][[index_col]].rename(columns={index_col:'区间分位最高30%'})
            a3 = index_risk[(index_risk[index_col_pct] <= (0.7 + diff))&(index_risk[index_col_pct] >= 0.3-diff)][[index_col]].rename(columns={index_col:'中间区间分位'})
            a4 = index_risk[(index_risk[index_col_pct] <= 0.3+diff)&(index_risk[index_col_pct] >= 0.05-diff)][[index_col]].rename(columns={index_col:'区间分位最低30%'})
            a5 = index_risk[(index_risk[index_col_pct] <= 0.1+diff)][[index_col]].rename(columns={index_col:'区间分位最低10%'})
            am = pd.DataFrame(index_risk[index_col].rolling(window=window,min_periods=min_periods).mean())
            am.columns=['均值']
            df = pd.concat([a1,a2,a3,a4,a5,am],axis=1).sort_index().dropna(axis=0,how='all')
            df.index.name = 'datetime'
            
            cols = df.columns.tolist()
            cols = [i for i in cols if i != '均值']
            for col in cols:
                df_i = df[[col]].copy()
                df_i.loc[:,'con'] = df_i[col].isnull()
                df_i.loc[:,'last_con'] = df_i[col].shift(-1).isnull()
                dts = df_i[( ~df_i.con) & (df_i.last_con)].index.tolist()
                for i in dts:
                    if df.loc[i:].shape[0] == 1:
                        continue
                    else:
                        i = df.loc[i:].index[1]
                        df.loc[i,col] = index_risk.risk_value.loc[i]
            
            title = index_name +title_dic[valuation_type]
            data = {
                'data':df,
                'title':title,
            }
            return data
        except Exception as e:
            print(f'failed to get data <err_msg> {e} from derived.stock_debt_val_detail')
            return False

    def industry_beta(self,begin_date, end_date, time_para):
        try:
            begin_date, end_date = RawDataApi().get_date_range(time_para, begin_date, end_date)
            index_info = BasicDataApi().get_asset_info_by_type(['行业指数'])
            index_ids = index_info.real_id.tolist()
            df = DerivedDataApi().get_index_valuation_develop(index_ids=index_ids,start_date=begin_date,end_date=end_date)
            pb_df = df.pivot_table(index='datetime',values='pb_mrq',columns='index_id')
            roe_df = df.pivot_table(index='datetime',values='roe',columns='index_id')   
            result = []
            dt_list = pb_df.index.tolist()
            for idx, dt, in enumerate(dt_list):
                if idx % 5 != 0:
                    continue
                _x = roe_df.loc[dt].dropna()
                x = sm.add_constant(_x)
                y = np.log(pb_df.loc[dt].dropna())
                model = regression.linear_model.OLS(y,x).fit()
                beta = model.params[1]
                result.append({'datetime':dt,'beta':beta})
            _df = pd.DataFrame(result).set_index('datetime')
            index_price = BasicDataApi().get_index_price_dt(start_date=_df.index[0],end_date=_df.index[-1],index_list=['hs300'])
            index_price = index_price.pivot_table(index='datetime',values='close',columns='index_id')
            _df = _df.join(index_price).ffill().reset_index()
            return _df
        except Exception as e:
            print(f'failed to get data <err_msg> {e} from derived.stock_debt_val_detail')
            return False




from typing import Optional, List, Dict, Any
import os
import traceback
import time
import datetime

from sqlalchemy.orm import sessionmaker
import pandas as pd
import numpy as np

from ...util.mail_retriever import MailAttachmentRetriever, UID_FILE_NAME
from ...util.wechat_bot import WechatBot
from ...util.calculator import Calculator
from ..wrapper.mysql import BasicDatabaseConnector, DerivedDatabaseConnector
from ..view.derived_models import FOFNav
from ..api.basic import BasicDataApi
from ..api.derived import DerivedDataApi
from ..view.basic_models import FOFInfo
from ..manager.manager_fof import FOFDataManager


class FOFNAVReader:

    COLUMNS_DICT = {
        '基金代码': 'fof_id',
        '产品代码': 'fof_id',
        '资产代码': 'fof_id',
        '基金名称': 'fund_name',
        '产品名称': 'fund_name',
        '资产名称': 'fund_name',
        '基金份额净值': 'nav',
        '单位净值': 'nav',
        '计提前单位净值': 'nav',
        '资产份额净值(元)': 'nav',
        '基金份额累计净值': 'acc_net_value',
        '累计单位净值': 'acc_net_value',
        '累计净值': 'acc_net_value',
        '资产份额累计净值(元)': 'acc_net_value',
        '虚拟后净值': 'v_net_value',
        '虚拟净值': 'v_net_value',
        '计提后单位净值': 'v_net_value',
        '虚拟计提净值': 'v_net_value',
        '虚拟后单位净值': 'v_net_value',
        '复权累计净值': 'adjusted_nav',
        '复权净值': 'adjusted_nav',
        '净值(分红再投)': 'adjusted_nav',
        '日期': 'datetime',
        '净值日期': 'datetime',
        '业务日期': 'datetime',
        '估值日期': 'datetime',
        '计算日期': 'calc_date',
    }

    def __init__(self, info: Dict[str, Any]):
        pd.set_option('display.max_rows', None)
        pd.set_option('display.max_columns', None)

        email_data_base_dir = os.environ['SURFING_EMAIL_DATA_DIR']

        self._read_dir = os.path.join(email_data_base_dir, f"attachments/{info['manager_id']}_{info['fof_id']}")
        os.makedirs(self._read_dir, exist_ok=True)
        assert os.path.isdir(self._read_dir), f'arg dump_dir should be a directory (now){self._read_dir}'

        self._manager_id = info['manager_id']
        self._fof_id = info['fof_id']
        self._sp_uri = f"{info['server']}:{info['port']}"
        self._user_name = info['email']
        self._password = info['password']
        self._wechat_bot = WechatBot()

    def _read_for_Template1(self, file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['估值日期', '单位净值', '累计单位净值']]
        df = df.rename(columns=FOFNAVReader.COLUMNS_DICT)
        df['fof_id'] = self._fof_id
        return df

    @staticmethod
    def _read_for_Template2(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品代码', '产品名称', '净值日期', '单位净值', '累计净值']]
        df = df.rename(columns=FOFNAVReader.COLUMNS_DICT)
        return df

    @staticmethod
    def _read_for_OrientSec(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path, header=None)
        df.iloc[:, 0] = df.iloc[:, 0].map(lambda x: x.strip().split('：')[0] if isinstance(x, str) else x)
        df = df.set_index(df.columns[0])
        date = df.index.array[2].date()
        df = df.T.loc[:, ['基金代码', '基金名称', '基金份额净值', '基金份额累计净值']]
        df = df[df.notna().any(axis=1)]
        df = df.rename(columns=FOFNAVReader.COLUMNS_DICT).assign(datetime=date)
        return df

    @staticmethod
    def _read_for_GuoTaiJunAnSec(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['产品代码', '产品名称', '净值日期', '单位净值', '累计单位净值']]
        df = df.rename(columns=FOFNAVReader.COLUMNS_DICT)
        return df

    @staticmethod
    def _read_for_HuaTaiSec(file_path: str) -> pd.DataFrame:
        df = pd.read_excel(file_path)
        df = df.loc[:, ['资产代码', '资产名称', '日期', '资产份额净值(元)', '资产份额累计净值(元)']]
        df = df.rename(columns=FOFNAVReader.COLUMNS_DICT)
        return df

    def _notify_error_event(self, err_msg: str):
        print(f'[read_navs_and_dump_to_db] {err_msg}')
        # self._wechat_bot.send_hedge_fund_nav_update_failed(err_msg)

    def read_navs_and_dump_to_db(self):
        fof_info = BasicDataApi().get_fof_info(self._manager_id, [self._fof_id])
        custodian_name = fof_info.custodian_name.array[0]
        try:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'rb') as f:
                uid_last = f.read()
                if not uid_last:
                    uid_last = None
        except FileNotFoundError:
            uid_last = None
        except Exception as e:
            self._notify_error_event(f'read uid file failed (e){e}, use None instead(read all emails) (manager_id){self._manager_id} (fof_id){self._fof_id}')
            uid_last = None

        try:
            mar = MailAttachmentRetriever(self._read_dir, ['xls', 'xlsx', 'pdf'])
            data = mar.get_excels(self._sp_uri, self._user_name, self._password, uid_last)
        except Exception as e:
            self._notify_error_event(f'FATAL ERROR!! get new data of hedge fund nav failed (e){e} (manager_id){self._manager_id} (fof_id){self._fof_id}')
            return

        parsers = {
            '东方证券': FOFNAVReader._read_for_OrientSec,
            '国泰君安证券': FOFNAVReader._read_for_GuoTaiJunAnSec,
            '中信证券': self._read_for_Template1,
            '中国国际金融股份': FOFNAVReader._read_for_Template2,
            '华泰证券': FOFNAVReader._read_for_HuaTaiSec,
        }

        can_not_update = False
        uid_last_succeed: Optional[bytes] = None
        df_list: List[pd.DataFrame] = []
        for name, comp_date in data.items():
            uid, file_path = comp_date

            try:
                for c_name, func in parsers.items():
                    if c_name in custodian_name:
                        try:
                            df = func(file_path)
                            print(df)
                        except Exception:
                            pass
                            # can_not_update = True
                        else:
                            break
                else:
                    raise NotImplementedError(f'unknown hedge fund nav file from attachment (manager_id){self._manager_id} (fof_id){self._fof_id}')
            except Exception as e:
                if not can_not_update:
                    # 走到这里都认为是已经处理完了这条数据
                    uid_last_succeed = uid
                self._notify_error_event(f'{e} (parse) (name){name} (file_path){file_path} (manager_id){self._manager_id} (fof_id){self._fof_id}')
                continue

            if not can_not_update:
                # 走到这里都认为是已经处理完了这条数据
                uid_last_succeed = uid

            try:
                df = df[df.fof_id == self._fof_id]
                if df.empty:
                    continue
                df = df.drop(columns=['fund_name'], errors='ignore')
                df['manager_id'] = self._manager_id
                df = self._dump_to_db(df)
                if df is not None:
                    df_list.append(df)
                else:
                    print(f'[read_navs_and_dump_to_db] duplicated data, do not process it (name){name} (manager_id){self._manager_id} (fof_id){self._fof_id}')
                time.sleep(1)
            except Exception as e:
                traceback.print_exc()
                self._notify_error_event(f'{e} (dump) (name){name} (file_path){file_path} (manager_id){self._manager_id} (fof_id){self._fof_id}')
                break

        if df_list:
            try:
                whole_df = pd.concat(df_list) #.set_index('fof_id')
                print(whole_df)
                # self._wechat_bot.send_hedge_fund_nav_update(whole_df)
            except Exception as e:
                self._notify_error_event(f'{e} (concat) (manager_id){self._manager_id} (fof_id){self._fof_id}')
                return
            else:
                print(f'[read_navs_and_dump_to_db] done (uid_last){uid_last_succeed} (df){whole_df} (manager_id){self._manager_id} (fof_id){self._fof_id}')
        else:
            whole_df = None
            print(f'[read_navs_and_dump_to_db] no new data this time, done (uid_last){uid_last_succeed} (manager_id){self._manager_id} (fof_id){self._fof_id}')
        # 记录下成功的最后一个uid
        if uid_last_succeed is not None:
            with open(os.path.join(self._read_dir, UID_FILE_NAME), 'wb') as f:
                f.write(uid_last_succeed)

            print(f'[read_navs_and_dump_to_db] to update fof info (manager_id){self._manager_id} (fof_id){self._fof_id}')
            fof_nav = DerivedDataApi().get_fof_nav(self._manager_id, [self._fof_id])
            if fof_nav is not None and not fof_nav.empty:
                fof_nav = fof_nav.drop(columns=['update_time', 'create_time', 'is_deleted']).set_index('datetime').sort_index()
                fof_latest_acc_nav = fof_nav.acc_net_value.array[-1]
                fof_latest_adjusted_nav = fof_nav.adjusted_nav.array[-1]
                nav_of_fof = fof_nav.nav
                nav = nav_of_fof.array[-1]
                total_ret = nav_of_fof.array[-1] / nav_of_fof.array[0] - 1
                res_status = Calculator.get_stat_result(nav_of_fof.index, nav_of_fof.array)
                last_year_nav = nav_of_fof[nav_of_fof.index < datetime.date(nav_of_fof.index.array[-1].year, 1, 1)]
                if not last_year_nav.empty:
                    ret_year_to_now = nav_of_fof.array[-1] / last_year_nav.array[-1] - 1
                else:
                    ret_year_to_now = np.nan

                Session = sessionmaker(BasicDatabaseConnector().get_engine())
                db_session = Session()
                fof_info_to_set = db_session.query(FOFInfo).filter((FOFInfo.manager_id == self._manager_id) & (FOFInfo.fof_id == self._fof_id)).one_or_none()
                fof_info_to_set.net_asset_value = float(nav) if not pd.isnull(nav) else None
                fof_info_to_set.acc_unit_value = float(fof_latest_acc_nav) if not pd.isnull(fof_latest_acc_nav) else None
                fof_info_to_set.adjusted_net_value = float(fof_latest_adjusted_nav) if not pd.isnull(fof_latest_adjusted_nav) else None
                # fof_info_to_set.total_volume = float(self._total_shares) if not pd.isnull(self._total_shares) else None
                # fof_info_to_set.total_amount = float(self._total_net_assets) if not pd.isnull(self._total_net_assets) else None
                fof_info_to_set.latest_cal_date = nav_of_fof.index.array[-1]
                fof_info_to_set.ret_year_to_now = float(ret_year_to_now) if not pd.isnull(ret_year_to_now) else None
                fof_info_to_set.ret_total = float(total_ret) if not pd.isnull(total_ret) else None
                fof_info_to_set.ret_ann = float(res_status.annualized_ret)
                fof_info_to_set.mdd = float(res_status.mdd)
                fof_info_to_set.sharpe = float(res_status.sharpe)
                fof_info_to_set.vol = float(res_status.annualized_vol)
                db_session.commit()
                db_session.close()

                if whole_df is not None:
                    print(f'[read_navs_and_dump_to_db] to calc and update adj net value (manager_id){self._manager_id} (fof_id){self._fof_id}')
                    adj_nav = FOFDataManager._calc_adj_nav_for_a_fund(fof_nav.reset_index().rename(columns={'fof_id': 'fund_id', 'nav': 'net_asset_value', 'acc_net_value': 'acc_unit_value'}))
                    if not adj_nav.empty and not pd.isnull(adj_nav.ta_factor):
                        print(adj_nav)
                        Session = sessionmaker(DerivedDatabaseConnector().get_engine())
                        db_session = Session()
                        hedge_fund_nav_to_set = db_session.query(FOFNav).filter(FOFNav.fof_id == self._fof_id, FOFNav.manager_id == self._manager_id, FOFNav.datetime == adj_nav.datetime).one_or_none()
                        hedge_fund_nav_to_set.ta_factor = adj_nav.ta_factor
                        hedge_fund_nav_to_set.adjusted_net_value = adj_nav.adj_nav
                        db_session.commit()
                        db_session.close()
            return whole_df
        return

    def _dump_to_db(self, df: pd.DataFrame):
        def _check_after_merged(x: pd.Series, now_df: pd.DataFrame):
            try:
                now_data = now_df[(now_df.fof_id == x.fof_id) & (now_df.datetime == x.datetime)]
                if now_data[['nav', 'acc_net_value', 'adjusted_nav']].iloc[0].astype('float64').equals(x[['nav', 'acc_net_value', 'adjusted_nav']].astype('float64')):
                    return pd.Series(dtype='object')
                else:
                    return x
            except (KeyError, IndexError):
                return x

        # TODO:
        assert df.datetime.nunique() == 1, 'should have single datetime'
        now_df = DerivedDataApi().get_fof_nav(self._manager_id, [self._fof_id])
        if now_df is not None and not now_df.empty:
            # 同产品同日期的净值如果已经存在了且没有变化，就不写DB了
            now_df = now_df.drop(columns=['update_time', 'create_time', 'is_deleted', 'volume', 'mv', 'ret', 'ta_factor']).sort_values(by=['fof_id', 'datetime']).drop_duplicates(subset=['fof_id', 'datetime'], keep='last')
            now_df = now_df.astype({'nav': 'float64', 'acc_net_value': 'float64', 'adjusted_nav': 'float64'})
            df = df.reindex(columns=now_df.columns).astype(now_df.dtypes.to_dict())
            df = df.merge(now_df, how='left', on=['manager_id', 'fof_id', 'datetime', 'nav', 'acc_net_value'], indicator=True, validate='one_to_one')
            df = df[df._merge == 'left_only'].drop(columns=['_merge', 'adjusted_nav_y']).rename(columns={'adjusted_nav_x': 'adjusted_nav'})
            if df.empty:
                return
            # FIXME 没想到特别好的方法 遍历每一行再check一下
            df['datetime'] = pd.to_datetime(df.datetime, infer_datetime_format=True).dt.date
            df = df.apply(_check_after_merged, axis=1, now_df=now_df)
            if df.empty:
                return
            df = df.set_index(['fof_id', 'datetime'])
            print(df)
            df.update(now_df.set_index(['fof_id', 'datetime']), overwrite=False)
            df = df.reset_index()
            df['datetime'] = df.datetime.map(lambda x: x.date())
            print(df)
            # 先删后添
            for date_to_delete in df.datetime.unique():
                DerivedDataApi().delete_fof_nav(date_to_delete=date_to_delete, manager_id=self._manager_id, fof_id_list=[self._fof_id])
        df.to_sql(FOFNav.__table__.name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')
        return df


if __name__ == '__main__':
    import requests

    url = 'https://fof.prism-advisor.com/api/v1/manager_mail/email_task'
    verify_token = 'jisn401f7ac837da42b97f613d789819f37bee6a'

    res = requests.post(
        url=url,
        data={
            'verify_token': verify_token,
        }
    )
    for one in res.json()['data']:
        if one['fof_id'] != 'SLW695':
            continue
        print(one)
        fof_nav_r = FOFNAVReader(one)
        fof_nav_r.read_navs_and_dump_to_db()




from datetime import datetime
import traceback

import pandas as pd

from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.basic_models import StyleAnalysisStockFactor
from .basic_data_helper import BasicDataHelper, add_column_to_style_analysis_factor


class StockFactor:

    _SHORT_TERM_DAYS = 20
    _LONG_TERM_DAYS = 242
    _HIGH_LOW_DAYS = 20

    def __init__(self, _data_helper):
        self.data_helper: BasicDataHelper = _data_helper
        # 记录因子值
        self._factors: pd.DataFrame = None

    @property
    def SHORT_TERM_DAYS(self):
        return self._SHORT_TERM_DAYS

    @property
    def LONG_TERM_DAYS(self):
        return self._LONG_TERM_DAYS

    @property
    def HIGH_LOW_DAYS(self):
        return self._HIGH_LOW_DAYS

    def init(self, start_date: str, end_date: str):
        # 由于有rolling, 所以需要从start_date往前多取一些数据
        trading_day = BasicDataApi().get_trading_day_list().set_index('datetime')
        start_date_loc = trading_day.index.get_loc(datetime.strptime(start_date, '%Y-%m-%d').date(), method='ffill')
        assert isinstance(start_date_loc, int), 'start date is not a valid trading day'

        extra_days = max((self.SHORT_TERM_DAYS, self.LONG_TERM_DAYS, self.HIGH_LOW_DAYS))
        assert start_date_loc >= extra_days, 'there is no enough extra days before start date'

        start_date_with_extra_days = trading_day.index.array[start_date_loc-extra_days]
        print(f'start date with extra days: {start_date_with_extra_days}')

        raw_api = RawDataApi()
        # 获取未复权收盘价
        self._stock_price = raw_api.get_em_stock_price(start_date_with_extra_days, end_date, columns=['close', 'trade_status']).set_index(['stock_id', 'datetime'])
        # 干掉当天没有进行交易的股票, 但这样的话往前取的数据可能不够rolling窗口使用了
        # self._stock_price[~self._stock_price['trade_status'].isin(('停牌一天', '暂停上市', '连续停牌'))]
        print('load stock price done')

        # 使用未复权股价的索引
        self._factors = pd.DataFrame(index=self._stock_price.index)

        # 获取后复权收盘价及前收
        self._stock_post_price = raw_api.get_em_stock_post_price(start_date_with_extra_days, end_date, columns=['close', 'pre_close', 'trade_status']).set_index(['stock_id', 'datetime'])
        # 干掉当天没有进行交易的股票, 但这样的话往前取的数据可能不够rolling窗口使用了
        # self._stock_post_price[~self._stock_post_price['trade_status'].isin(('停牌一天', '暂停上市', '连续停牌'))]
        # 按道理索引应该与未复权df完全一致，但可能值是一样的但顺序不同，这里reindex一下
        self._stock_post_price = self._stock_post_price.reindex(index=self._stock_price.index)
        print('load stock post price done')
        pd.testing.assert_index_equal(self._stock_post_price.index, self._stock_price.index)
        # reindex可能掩盖了不一致，这里看一下是否有某一行除了索引之外全为null
        assert not self._stock_post_price.isnull().all(axis=1).any(), 'stock post price df should not have any NaN!!'

        # 获取每日更新的总股本
        self._stock_daily_info = raw_api.get_em_daily_info(start_date_with_extra_days, end_date, columns=['total_share']).set_index(['stock_id', 'datetime'])
        # 与未复权股价的索引对齐(空值保持为nan)
        self._stock_daily_info = self._stock_daily_info.reindex(index=self._stock_price.index)
        # 对齐后将每只个股时间序列上总股本为nan的值ffill, TODO 再bfill
        self._stock_daily_info['total_share'] = self._stock_daily_info.reset_index().pivot_table(index='stock_id', columns='datetime', values='total_share', dropna=False).fillna(method='ffill', axis=1).fillna(method='bfill', axis=1).stack()
        print('load stock daily info done')
        assert not self._stock_daily_info.isnull().to_numpy().any(), 'stock daily info df should not have any NaN!!'

        # 获取归属于母公司股东权益合计以及其他权益工具(计算PB)
        # 注意这里不传日期参数, 获取全表的数据; 这是必要的, 因为下边的ffill需要一直向上来找到有效的值
        self._stock_fin_fac = raw_api.get_em_stock_fin_fac().set_index(['stock_id', 'datetime']).loc[:, ['balance_statement_140', 'balance_statement_195']]
        # 先做一个fillna让所有财报日期上都有值, 因为后续reindex会让很多日期所对应的行消失
        self._stock_fin_fac['balance_statement_140'] = self._stock_fin_fac.reset_index().pivot_table(index='stock_id', columns='datetime', values='balance_statement_140', dropna=False).fillna(method='ffill', axis=1).stack()
        # 与未复权股价的索引对齐(空值保持为nan)
        self._stock_fin_fac = self._stock_fin_fac.reindex(index=self._stock_price.index)
        # 其他权益工具的nan均置为0
        self._stock_fin_fac['balance_statement_195'] = self._stock_fin_fac['balance_statement_195'].fillna(0)
        # 对齐后将每只个股时间序列上归属于母公司股东权益合计为nan的值ffill, 因此这里start_date_with_extra_days应该是一个财报日期（最好为年报）
        # TODO: 先再做个bfill
        self._stock_fin_fac['balance_statement_140'] = self._stock_fin_fac.reset_index().pivot_table(index='stock_id', columns='datetime', values='balance_statement_140', dropna=False).fillna(method='ffill', axis=1).fillna(method='bfill', axis=1).stack()
        print('load stock fin fac done')

    def _lambda_1(self, x) -> pd.Series:
        window = x.rolling(window=self.HIGH_LOW_DAYS)
        return (window.max() - window.min()) / (window.max() + window.min())

    def calculate_factors(self):
        # 收益率 = 每日收盘价 / 每日前收盘价 - 1, 后复权
        self._factors['rate_of_return'] = self._stock_post_price.loc[:, 'close'] / self._stock_post_price.loc[:, 'pre_close'] - 1
        print('calculate rate_of_return done')

        # 规模 = 每日总股本 * 每日收盘价, 未复权
        self._factors['latest_size'] = self._stock_daily_info.loc[:, 'total_share'] * self._stock_price.loc[:, 'close']
        print('calculate latest_size done')

        # 价值 = 1 / PB = (归属于母公司股东权益合计 - 其他权益工具) / 规模, 未复权
        self._factors['bp'] = (self._stock_fin_fac.loc[:, 'balance_statement_140'] - self._stock_fin_fac.loc[:, 'balance_statement_195']) / self._factors.loc[:, 'latest_size']
        print('calculate bp done')

        # 短期动量, group by stock_id, 在每一只股票上按时间做rolling, 取窗口内日收益率的和
        self._factors['short_term_momentum'] = self._factors.loc[:, ['rate_of_return']].groupby(level='stock_id').apply(lambda x: x.rolling(window=self.SHORT_TERM_DAYS).sum())
        print('calculate short_term_momentum done')

        # 长期动量, group by stock_id, 在每一只股票上按时间做rolling, 取窗口内日收益率的和
        self._factors['long_term_momentum'] = self._factors.loc[:, ['rate_of_return']].groupby(level='stock_id').apply(lambda x: x.rolling(window=self.LONG_TERM_DAYS).sum())
        print('calculate long_term_momentum done')

        # 波动率, group by stock_id, 在每一只股票上按时间做rolling, 取窗口内收盘价(最大值-最小值)/(最大值+最小值)
        self._factors['high_low'] = self._stock_post_price.loc[:, ['close']].groupby(level='stock_id').apply(self._lambda_1)
        print('calculate high_low done')

    def process_all(self, start_date: str, end_date: str):
        start_date = datetime.strptime(start_date, '%Y%m%d').date().isoformat()
        end_date = datetime.strptime(end_date, '%Y%m%d').date().isoformat()
        failed_tasks = []
        try:
            self.init(start_date, end_date)
            self.calculate_factors()
            self.data_helper._upload_basic(self._factors.reset_index().set_index('datetime').loc[start_date:end_date].reset_index(), StyleAnalysisStockFactor.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(StyleAnalysisStockFactor.__table__.name)
        return failed_tasks

    # 注意该函数应该一次性执行后将一些逻辑加到process_all函数中, 以实现新因子值的每日更新
    def add_factor(self):
        # 先获取DB中整张数据表的日期范围, 以此来作为新因子计算时所需数据的日期范围
        time_range_df = BasicDataApi().get_style_analysis_time_range()
        assert time_range_df.shape[0] == 1, 'invalid time range df'

        start_date, end_date = time_range_df.iloc[0, :].loc[['start_date', 'end_date']].to_list()
        print(f'time range for adding factor: {start_date} to {end_date}')

        # 通过api获取计算新因子所需的数据

        # 计算新因子每一天的值

        # 将新因子插入到DB中新的一列
        pivot_df = pd.DataFrame()
        new_column_name = ''
        add_column_to_style_analysis_factor(pivot_df, new_column_name)


if __name__ == "__main__":
    StockFactor(BasicDataHelper()).process_all('20031231', '20141231')



import pandas as pd
import traceback
from ...wrapper.mysql import DerivedDatabaseConnector, ViewDatabaseConnector, BasicDatabaseConnector
from ...view.view_models import AutomaticInvestmentCollection
from ...view.derived_models import FundAIC
from ...view.basic_models import FundInfo
class AICDailyCollectionProcessor(object):

    def get_fund_df(self):
        with BasicDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    FundInfo.fund_id,
                    FundInfo.structure_type
                )
                funds = pd.read_sql(query.statement, query.session.bind)
                return funds
            except Exception as e:
                print('Failed fill get data <err_msg> {}'.format(e))

    def get_fund_aic(self):
        # 只选择非分级基金和分级基金母基金
        fund_info = self.get_fund_df()
        fund_list = fund_info[fund_info.structure_type <= 1].fund_id.tolist()
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                latest_time = mn_session.query(FundAIC).order_by(FundAIC.datetime.desc()).limit(1).one_or_none()
                latest_time = latest_time.datetime
                print(latest_time)
                query = mn_session.query(
                    FundAIC.fund_id,
                    FundAIC.datetime,
                    FundAIC.y1_ret,
                    FundAIC.y3_ret,
                    FundAIC.y5_ret,
                    FundAIC.intel_y1_ret,
                    FundAIC.intel_y3_ret,
                    FundAIC.intel_y5_ret,
                ).filter(FundAIC.datetime==latest_time,
                        FundAIC.fund_id.in_(fund_list))
                df = pd.read_sql(query.statement, query.session.bind)
                df = df.rename(columns={
                    'datetime': 'price_time',
                })
                return df
            except Exception as e:
                print('Failed get_index_volatility <err_msg> {}'.format(e))

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def collection_daily_aic(self):
        try:
            print('1、 load data...')
            df = self.get_fund_aic()

            self.append_data(AutomaticInvestmentCollection.__tablename__, df)
            print(df)
            print('AIC success')
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def process(self):
        failed_tasks = []
        if not self.collection_daily_aic():
            failed_tasks.append('collection_daily_aic')
        return failed_tasks


if __name__ == '__main__':
    AICDailyCollectionProcessor().collection_daily_aic()



from .message_sender import MessageSender
import datetime
import requests
import json
import time

class TextMessageSender(MessageSender):

    def __init__(self):
        super(TextMessageSender, self).__init__()

    def send(self, msg_to, msg_content):
        # Send snapshot of webpage and url
        # doc: https://open.work.weixin.qq.com/api/doc/90000/90135/90236
        access_token = self.get_access_token()
        target_url = f'https://api.weixin.qq.com/cgi-bin/message/custom/send?access_token={access_token}'

        # Send url
        message = {
            'touser': '|'.join(msg_to) if type(msg_to) is list else msg_to,
            'msgtype': 'text',
            'text': {'content': msg_content}
        }
        print(message)
        res = requests.post(url=target_url, data=json.dumps(message, ensure_ascii=False).encode(), timeout=20)

        # # For QR test
        # qr_url = f'https://api.weixin.qq.com/cgi-bin/qrcode/create?access_token={access_token}'
        # qr_message = {
        #     'expire_seconds': 1800, # QR is valid for 30 minutes
        #     'action_name': 'QR_SCENE',
        #     'action_info': {
        #         'scene': {
        #             'scene_id': 1122
        #         }
        #     }
        # }
        # print(qr_message)
        # res = requests.post(url=qr_url, data=json.dumps(qr_message, ensure_ascii=False).encode(), timeout=20)
        # print(res.text)



import pandas as pd
import numpy as np
import platform
import matplotlib as mpl
import pylab as pl
import datetime
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.dates as mdates
from ....util.calculator import Calculator
from matplotlib.patches import Rectangle
CURRENT_PLATFORM = platform.system()
if CURRENT_PLATFORM == 'Darwin':
    mpl.rcParams['font.family'] = ['Heiti TC']
else:
    mpl.rcParams['font.family'] = ['STKaiti']
    
pd.plotting.register_matplotlib_converters()
class HybridPainter(object):

    @staticmethod
    def plot_market_value(mv_df:pd.DataFrame, backtest_type:str, index_price:pd.DataFrame, saa_weight:dict, bk_stats:dict):
        index_df_raw = index_price.loc[mv_df.index[0]:mv_df.index[-1],]
        index_df = index_df_raw.copy().fillna(method='bfill')
        index_df = index_df / index_df.iloc[0]
        index_df['cash'] = 1
        w_l = []
        for idx, r in index_df_raw.iterrows():
            nan_asset = [k for k, v in r.to_dict().items() if np.isnan(v)]
            wgts = saa_weight.copy()
            for k in nan_asset:
                wgts[k] = 0
            s = sum([v  for k,v in wgts.items()])
            wgts = {k :v /s for k, v in wgts.items()}
            wgts['datetime'] = idx
            w_l.append(wgts)
        wgts_df = pd.DataFrame(w_l).set_index('datetime')
        mv_df['benchmark'] = (wgts_df * index_df).sum(axis = 1)
        mv_df = mv_df / mv_df.iloc[0]
        mv_df.plot.line(figsize=(20,12),legend=False,fontsize = 17)
        l = pl.legend(loc='lower left',fontsize = 17)
        s = pl.title(f'{backtest_type} market value', fontsize=20)
        ar = round(bk_stats['annual_ret'],4)
        mdd = round(bk_stats['mdd'], 4)
        sharpe = round(bk_stats['sharpe'],4)
        pl.suptitle(f'annual_ret : {ar} mdd : {mdd} sharpe : {sharpe}',y=0.87,fontsize=17)
        plt.grid()

    @staticmethod
    def plot_market_value_with_mdd_analysis(mv_df:pd.DataFrame, backtest_type:str, index_price:pd.DataFrame, saa_weight:dict, bk_stats:dict):
        def get_mdd_result_from_df(df:pd.DataFrame,
                                    date_column: str,
                                    value_column: str):
            dates = df[date_column].values
            values = df[value_column].values
            return mdd_recover_analysis(values,dates)

        def mdd_recover_analysis(values,dates):
            sr = pd.Series(values, index=dates).sort_index()
            if sr.empty:
                mdd = 0
                mdd_date1 = None
                mdd_date2 = None
                mdd_lens = 0
                return mdd, mdd_date1, mdd_date2, mdd_lens
            mdd_part =  sr[:] / sr[:].rolling(window=sr.shape[0], min_periods=1).max()
            mdd = 1 - mdd_part.min()
            if mdd == 0:
                mdd_date1 = None
                mdd_date2 = None
                mdd_lens = 0
            else:
                mdd_date = mdd_part.idxmin()
                mdd_date1 = sr[:mdd_date].idxmax()
                sr_tmp = sr[mdd_date1:]
                recover_sr = sr_tmp[sr_tmp> sr[mdd_date1]]
                if recover_sr.empty:
                    mdd_date2 = sr_tmp.index[-1]
                else: 
                    mdd_date2 = sr_tmp[sr_tmp> sr[mdd_date1]].index[0]
                mdd_lens = sr.loc[mdd_date1:mdd_date2].shape[0]
            return mdd, mdd_date1, mdd_date2, mdd_lens
                
        def plot_mdd_recover(mdd_date1_1, mdd_date1_2, df, c, ax):
            start = mdates.date2num(mdd_date1_1)
            end = mdates.date2num(mdd_date1_2)
            width = end - start
            price_max = df.loc[mdd_date1_1: mdd_date1_2].mv.max()
            bottom = df.loc[mdd_date1_1: mdd_date1_2].mv.min()
            height = price_max - bottom
            rect = Rectangle((start, bottom), width, height, color=c)
            ax.add_patch(rect)   

        def print_st(mdd):
            return str(round(mdd * 100,2)) + '%' 
            
        
        mdd1, mdd_date1_1, mdd_date1_2, mdd_lens1 = get_mdd_result_from_df(mv_df.reset_index(),'date','mv')
        mv_df_1 = mv_df.loc[:mdd_date1_1].reset_index()
        mv_df_2 = mv_df.loc[mdd_date1_2:].reset_index()

        mdd2, mdd_date2_1, mdd_date2_2, mdd_lens2 = get_mdd_result_from_df(mv_df_1,'date','mv')
        mdd3, mdd_date3_1, mdd_date3_2, mdd_lens3 = get_mdd_result_from_df(mv_df_2,'date','mv')
        if mdd3 > mdd2:
            mdd2 = mdd3
            mdd_date2_1 = mdd_date3_1
            mdd_date2_2 = mdd_date3_2
            mdd_lens2 = mdd_lens3
             
        index_df_raw = index_price.loc[mv_df.index[0]:mv_df.index[-1],]
        date_list = [datetime.datetime.strptime(i, '%Y-%m-%d').date() for i in bk_stats['hold_years']]
        date_list = sorted(list(set(date_list + [mv_df.index[-1]])))
        saa_weight = { k : v for k, v in saa_weight.items() if v > 0}
        _res = []
        last_value = 1
        for dt in date_list[:-1]:
            next_dt = date_list[date_list.index(dt) + 1]
            _df = index_df_raw.loc[dt:next_dt][list(saa_weight.keys())]
            _df = _df / _df.iloc[0]
            res_i = (_df * saa_weight).sum(axis=1)
            res_i = res_i * last_value
            last_value = res_i.values[-1]
            _res.append(res_i)
        benchmark_df = pd.concat(_res,axis=0)
        mv_df = mv_df.join(pd.DataFrame(benchmark_df,columns=['benchmark'])).bfill()
        mv_df = mv_df.drop_duplicates()
        df = mv_df / mv_df.iloc[0]
        df.index.name = 'date'
        fig, ax = plt.subplots(figsize= [20,12])
        df['price_ratio'] = df['mv'] / df['benchmark']       
        plt.plot(df.index, df['mv'], label = 'mv', linewidth=2.0)
        plt.plot(df.index, df['benchmark'], label = 'benchmark', linewidth=2.0)
        plt.plot(df.index, df['price_ratio'], label = 'price ratio', linewidth=2.0)
        
        stats_res = Calculator.get_stat_result_from_df(df=df.reset_index(), date_column='date', value_column='price_ratio').__dict__
        ar = print_st(bk_stats['annual_ret'])
        av = print_st(bk_stats['annual_vol'])
        sharpe = round(bk_stats['sharpe'],2)
        st1 = f'annual_ret : {ar} annual_vol : {av} sharpe : {sharpe}'
        st2 = f'first mdd {print_st(mdd1)} {mdd_date1_1} : {mdd_date1_2}, {mdd_lens1} days'
        st3 = f'second mdd {print_st(mdd2)} {mdd_date2_1} : {mdd_date2_2}, {mdd_lens2} days'
        ar_ = print_st(stats_res['annualized_ret'])
        av_ = print_st(stats_res['annualized_vol'])
        mdd_ = print_st(stats_res['mdd'])
        sharpe_ = round(stats_res['sharpe'],2)
        st4 = f'price ratio annaul ret {ar_} annual vol {av_} mdd {mdd_} sharpe {sharpe_}'
        st = st1 + '\n' + st2 + '\n' + st3 + '\n' + st4
        plt.legend(fontsize=20 ,loc = 'upper left')
        plt.title(f'market values ', fontsize=30)
        plt.suptitle(st, y=0.87, fontsize=18)
        plot_mdd_recover(mdd_date1_1, mdd_date1_2, df, 'silver', ax)
        plot_mdd_recover(mdd_date2_1, mdd_date2_2, df, 'gainsboro', ax)
        plt.grid() 
        #plt.savefig("test.svg")
        plt.show()

    @staticmethod
    def plot_asset_fund_mv_diff(asset_mv:pd.DataFrame, fund_mv:pd.DataFrame):
        asset_mv = asset_mv[['mv']]
        fund_mv = fund_mv[['mv']]
        asset_mv.columns = ['asset_mv']
        fund_mv.columns = ['fund_mv']
        check_diff = fund_mv.join(asset_mv)
        check_diff = check_diff / check_diff.iloc[0]
        check_diff['diff'] = 100 * (check_diff['fund_mv']  - check_diff['asset_mv']) / check_diff['asset_mv'] 
        check_diff[['fund_mv','asset_mv']].plot.line(figsize=(20,12),legend=False,fontsize = 17)
        l = pl.legend(loc='lower left',fontsize = 17)
        s = pl.title('asset and fund market value ', fontsize=20)
        plt.grid()
        check_diff[['diff']].plot.line(figsize=(20,12),legend=False,fontsize = 17)
        l = pl.legend(loc='lower left',fontsize = 17)
        s = pl.title('asset and fund market value diff % ', fontsize=20)
        plt.grid() 
        plt.show()

    @staticmethod
    def plot_four_layers_mv(mv_df:pd.DataFrame):
        mv_df = mv_df / mv_df.iloc[0]
        mv_df.plot.line(color=['C3','C1','C2','C0'],figsize=(14,8))
        plt.title('净值对比', fontsize=20)
        plt.legend(loc='upper left', fontsize=15)
        plt.grid() 
        plt.show()

    @staticmethod
    def plot_taa_detail(df, title):
        df.plot(figsize=(14,8))
        _res = []
        for index_id, v in df.iloc[-1].to_dict().items():
            _res.append(f'{index_id}:{round(v,3)}')
        s = ' '.join(_res)
        plt.title(title, fontsize=25)
        plt.legend(loc='lower left', fontsize=15)
        plt.suptitle(s,y=0.87,fontsize=17)
        plt.grid()
        plt.show()

    @staticmethod
    def plot_asset_price_ratio(fund_nav:pd.DataFrame, 
                        fund_info:pd.DataFrame, 
                        index_price:pd.DataFrame,
                        score_dict:dict,
                        backtest_result:dict,
                        dts:pd.Series,
                        saa:dict,
                        fund_trade: pd.DataFrame,
                        score_select: dict,
                        ):
        
        proper_number_dict = {'mmf':1} 
        for asset, w in saa.items():
            if asset in ['cash','mmf'] or w == 0:
                continue
            num = min(int(round(w / 0.05, 0)) , 4)
            proper_number_dict[asset] = num
        
        fund_info['trade_fee'] = 0.15 * fund_info['purchase_fee'] + fund_info['redeem_fee']
        fee_df = fund_info[['fund_id','trade_fee']].set_index('fund_id')
        dts = dts[(dts >=fund_nav.index[0]) & (dts <= fund_nav.index[-1])]
        fund_nav = fund_nav.reindex(dts)
        index_price = index_price.reindex(dts)
        rebalance_date = backtest_result['rebalance_date'] + [backtest_result['market_value'].index.tolist()[-1]]
        fund_desc_name = fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']  
        res = {}
        for idx, begin_date in enumerate(rebalance_date[:-1]):
            end_date = rebalance_date[idx + 1]
            for asset, num in proper_number_dict.items():
                #print(asset, num)
                _score_dict = score_dict[begin_date][asset][score_select[asset]]
                _score_list = sorted(_score_dict.items(), key=lambda x:x[1], reverse=True)
                #_fund_list = fund_trade[(fund_trade['trade_date'] == begin_date) & (fund_trade['is_buy']) & (fund_trade['index_id'] == asset)].fund_id.tolist()
                _fund_list = [i[0] for i in _score_list]
                dic = {
                    'index_id':asset,
                    'begin_date':begin_date,
                    'end_date':end_date,
                    'fund_list':_fund_list[:proper_number_dict[asset]]
                }
                if asset in res:
                    res[asset].append(dic)
                else:
                    res[asset] = [dic]
        _res = []
        for index_id in res:
            _index_id_res = []
            for date_dic in res[index_id]:
                fund_nav_i = fund_nav.loc[date_dic['begin_date']:date_dic['end_date'],date_dic['fund_list']]
                for fund_id in fund_nav_i:
                    fund_i_fee = fee_df.loc[fund_id,'trade_fee']
                    fee_discount_l = np.linspace(1,1-fund_i_fee,fund_nav_i.shape[0])
                    fund_nav_i[fund_id] = fund_nav_i[fund_id] * fee_discount_l
                fund_ret_combine = pd.DataFrame(fund_nav_i.pct_change(1).mean(axis=1).dropna())
                _index_id_res.append(fund_ret_combine)
            fund_combine_rets = pd.concat(_index_id_res).rename(columns={0:'funds'})
            index_ret = index_price[[index_id]].loc[rebalance_date[0]: rebalance_date[-1]].pct_change(1).dropna()
            price_ratio_index = (fund_combine_rets.join(index_ret)+1).cumprod()
            price_ratio_index = pd.DataFrame(price_ratio_index['funds'] / price_ratio_index[index_id]).rename(columns={0:index_id})
            _res.append(price_ratio_index)
        df_plot = pd.concat(_res, axis=1, sort=False)
        res = []
        for index_id in df_plot:
            _index_stats = Calculator.get_stat_result_from_df(df=df_plot.reset_index(), date_column='datetime', value_column=index_id).__dict__
            _index_stats['index_id'] = index_id 
            res.append(_index_stats)
        
        stats = pd.DataFrame(res).set_index('index_id')[['annualized_ret','annualized_vol','mdd','sharpe','mdd_date1','mdd_date2']]
        df_plot.plot.line(figsize=(14,8))
        plt.title('基金所属各资产价格比',fontsize=20)
        plt.legend(loc='upper left', fontsize=15)
        plt.grid()
        plt.show()
        return stats

    @staticmethod
    def regular_rebalance_in_index_strategy(begin_date:datetime.date,
                                            end_date:datetime.date,
                                            index_id:str,
                                            rebalance_year:float,
                                            top_funds:int,
                                            dts:pd.Series,
                                            fund_score_dic:dict,
                                            fund_info:pd.DataFrame,
                                            index_price:pd.DataFrame,
                                            fund_nav:pd.DataFrame):
        
        def _get_top_funds(fund_score_dic, index_id, dt, top_nums):
            _score_dict = fund_score_dic[dt][index_id]
            _score_list = sorted(_score_dict.items(), key=lambda x:x[1], reverse=True)
            _fund_list = [i[0] for i in _score_list]
            if top_nums > len(_fund_list):
                return _fund_list
            else:
                return _fund_list[:top_nums]

        rebalance_date = int(rebalance_year * 242)
        fund_info['trade_fee'] = 0.15 * fund_info['purchase_fee'] + fund_info['redeem_fee']
        fee_df = fund_info[['fund_id','trade_fee']].set_index('fund_id')
        desc_name = fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']

        _dts = dts[(dts >= begin_date) & (dts <= end_date )].tolist()
        res = []
        for _idx, _dt in enumerate(_dts):
            if _idx % rebalance_date == 0:
                dic = {
                    'start_date':_dt,
                    'fund_list':_get_top_funds(fund_score_dic, index_id, _dt, top_funds),
                }
            elif _idx % rebalance_date == (rebalance_date-1):
                dic['end_date'] = _dt
                res.append(dic)
                dic = {
                    'start_date':_dt,
                    'fund_list':_get_top_funds(fund_score_dic, index_id, _dt, top_funds),
                }
        dic['end_date'] = _dt
        res.append(dic)
        _res = []
        for date_dic in res:
            fund_nav_i = fund_nav.loc[date_dic['start_date']:date_dic['end_date'],date_dic['fund_list']]
            for fund_id in fund_nav_i:
                fund_i_fee = fee_df.loc[fund_id,'trade_fee']
                fee_discount_l = np.linspace(1,1-fund_i_fee,fund_nav_i.shape[0])
                fund_nav_i[fund_id] = fund_nav_i[fund_id] * fee_discount_l
            fund_ret_combine = pd.DataFrame(fund_nav_i.pct_change(1).mean(axis=1).dropna())
            _res.append(fund_ret_combine)
        fund_combine_rets = pd.concat(_res).rename(columns={0:'funds'})
        fund_df = (fund_combine_rets+1).cumprod()
        fund_df.loc[_dts[0],'funds'] = 1
        fund_df = fund_df.sort_index()
        fund_df = fund_df.join(index_price[[index_id]]).fillna(method='ffill')
        fund_df = fund_df/fund_df.iloc[0]
        fund_df['price_ratio'] = fund_df['funds'] / fund_df[f'{index_id}']
        fund_df.plot.line(figsize=(15,8))
        plt.title(f'Selects top {top_funds} funds and rebalance every {rebalance_year} year in {index_id}',fontsize=25)
        plt.legend(loc='upper left', fontsize=15)
        plt.grid()
        plt.show()
        stats = Calculator.get_stat_result_from_df(df=fund_df.reset_index(), date_column='datetime', value_column='funds').__dict__
        result_df = pd.DataFrame([stats])[['annualized_ret','annualized_vol','mdd','sharpe','mdd_date1','mdd_date2']]
        result_df.index = ['portfolio result']
        _res = []
        for dic in res:
            _dic = {
                'start_date': dic['start_date'],
                'end_date':dic['end_date'],
            }
            for idx, _fund_id in enumerate(dic['fund_list']):
                _dic[f'fund_{idx+1}'] = desc_name[_fund_id]
            _res.append(_dic)
        pos_df = pd.DataFrame(_res)
        _col_list = pos_df.columns.tolist()
        _col_list1 = ['start_date','end_date']
        _col_list2 = [_ for _ in _col_list if _ not in _col_list1]
        col_list = _col_list1 + _col_list2
        pos_df = pos_df[col_list]
        return result_df, pos_df

    @staticmethod
    def fund_rank_recent_analysis(fund_id,index_id,score_dict,fund_info,fund_indicator,fund_nav,index_price,score_func,score_select):
        _res = []
        indicators = []
        for k,v in score_func.__dict__.items():
            if v !=0 :
                _res.append(f'{k} {round(v,2)}')
                indicators.append(k)
        _score_func = ' '.join(_res)  
        desc_name = fund_info[fund_info['fund_id'] == fund_id].desc_name.values[0]
        dts_list = sorted(list(score_dict.keys()))
        res = {}
        for dt in dts_list:
            _score_dict = score_dict[dt]
            if index_id in _score_dict:
                _score_dict = _score_dict[index_id][score_select[index_id]]
            else:
                continue
            _score_list = sorted(_score_dict.items(), key=lambda x:x[1], reverse=True)
            _fund_list = [i[0] for i in _score_list]
            res[dt] = _fund_list
        _res = []
        for dt, _fl in res.items():
            if fund_id in _fl:
                _res.append({'datetime':dt,'rank':_fl.index(fund_id)})
        rank_df = pd.DataFrame(_res)
        rank_df.set_index('datetime').plot(figsize=(16,8))
        plt.title(f'fund rank {desc_name} {fund_id}', fontsize=25)
        plt.legend(loc='upper right', fontsize=15)
        _res = []
        for i in rank_df.tail().to_dict('records'):
            _dt = i['datetime']
            _rank = i['rank']
            _res.append(f'{_dt} {_rank}')
        s = _score_func + '\n'+ '\n'.join(_res)

        plt.suptitle(s, y=0.87, fontsize=18)
        plt.show()

        dl = rank_df.tail().datetime.tolist()
        result = fund_indicator[(fund_indicator['fund_id'] == fund_id)
               &(fund_indicator['datetime'].isin(dl))][indicators+['datetime']].set_index('datetime')

        df = fund_nav[[fund_id]].join(index_price[index_id]).dropna()
        df = df / df.iloc[0]
        df.loc[:,'price_ratio'] = df[fund_id] / df[index_id]
        l1 = 3 * 242 + 10 if index_id in ['hs300','csi500'] else 242 + 10
        l = min(df.shape[0], l1)

        df.iloc[-l:]['price_ratio'].plot(figsize=(16,8))
        plt.title(f'{fund_id} {desc_name} indicator periods price_ratio', fontsize=25)
        plt.grid()
        plt.show()
        df['price_ratio'].plot(figsize=(16,8))
        plt.title(f'{fund_id} {desc_name} price_ratio', fontsize=25)
        plt.grid()
        plt.show()
        return result



from typing import List, Dict
import pandas as pd
import numpy as np
import math
import datetime
import traceback
import json
import statsmodels.api as sm
import time
from scipy.optimize import Bounds, minimize
from statsmodels.tsa.ar_model import AutoReg
from pandas.tseries.offsets import DateOffset
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.derived_models import FundIndicatorGroup
from .derived_data_helper import DerivedDataHelper


class FundIndicatorProcessorGroup:

    _TRADING_DAYS_PER_YEAR = 242
    _NATURAL_DAYS_PER_YEAR = 360
    _RISK_FEE_RATE = 0.011
    _RISK_FEE_RATE_PER_DAY = _RISK_FEE_RATE / _TRADING_DAYS_PER_YEAR
    _REPORT_DATE_LIST = ['0331', '0630', '0930', '1231']
    _FUND_CLASSIFIER = {
        'QDII': ['国际(QDII)股票型基金', '国际(QDII)债券型基金', '国际(QDII)另类投资基金', '国际(QDII)混合型基金', '股票多空', '商品型基金', 'REITs'],
        'stock': ['偏股混合型基金', '普通股票型基金', '被动指数型基金', '增强指数型基金'], 
        'bond': ['增强指数型债券基金','混合债券型二级基金', '中长期纯债型基金', '混合债券型一级基金', '短期纯债型基金', '偏债混合型基金', '被动指数型债券基金'],
        'mmf': ['货币市场型基金'],
        'mix': ['平衡混合型基金', '灵活配置型基金'],
    }
    _REPLACE_DICT = {'stock': 'csi_stockfund', 'bond': 'national_debt', 'mmf': 'mmf', 'QDII': 'csi_f_qdii', 'mix': 'csi_f_hybrid'}

    def __init__(self, data_helper: DerivedDataHelper, data_cycle: int):
        self._data_helper = data_helper
        # 单位: 年
        self._data_cycle = data_cycle
        self._data_cycle_days = data_cycle * self._TRADING_DAYS_PER_YEAR
        self._basic_api = BasicDataApi()
        self._rank_dic: Dict[pd.DataFrame] = {}
        self._res: List[pd.Series] = []

    def init(self, end_date: str, is_print_time: bool=False):
        # self.log = []
        self._wind_type_dict = {}
        for type_i, type_list in self._FUND_CLASSIFIER.items():
            for _ in type_list:
                self._wind_type_dict.update({_:type_i})
        # 获取待计算的基金列表
        self.fund_info: pd.DataFrame = self._basic_api.get_fund_info().drop(columns='_update_time')
        self._wind_class_2_dict = self.fund_info.set_index('fund_id')['wind_class_2'].to_dict()
        
        # 向前取N年
        start_date: str = (pd.to_datetime(end_date, infer_datetime_format=True) - DateOffset(years=self._data_cycle+1)).strftime('%Y%m%d')
        # 获取区间内交易日列表
        all_trading_days: pd.Series = self._basic_api.get_trading_day_list().drop(columns='_update_time').set_index('datetime')
        self._trading_days: pd.Series = all_trading_days.loc[pd.to_datetime(start_date, infer_datetime_format=True).date():pd.to_datetime(end_date, infer_datetime_format=True).date()]
        # 使用实际trading day的最后一天作为end_date
        self.end_date = self._trading_days.index.array[-1]

        # 基金净值数据、指数价格数据的index对齐到交易日列表
        _tm_loads_nav_0 = time.time()
        self._fund_list = self.fund_info[(self.fund_info.end_date > self.end_date) 
                                       & (self.fund_info.structure_type <= 1)
                                       & (~self.fund_info.wind_class_2.isnull())]
        fund_nav: pd.DataFrame = self._basic_api.get_fund_nav_with_date(start_date, self.end_date, self._fund_list.fund_id.tolist())
        self._fund_nav_origin: pd.DataFrame = fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').reindex(self._trading_days.index).fillna(method='ffill')
        _tm_loads_nav_1 = time.time()
        # 选取成立满半年的 而且近半年波动率不为0，并排除掉异常净值基金
        self.one_day_jump = self._fund_nav_origin.pct_change(1).abs().max()
        self.three_day_jump = self._fund_nav_origin.pct_change(3).abs().max()
        normal_fund_list = []
        self.abnormal_reason = []
        for fund_id in self._fund_nav_origin:
            _df = self._fund_nav_origin[fund_id].dropna()
            if _df.shape[0] > 120:
                if _df.iloc[-120:].var() > 1e-6:
                    wind_class_2 = self._wind_class_2_dict[fund_id]
                    select_type = self._wind_type_dict.get(wind_class_2, None)
                    if select_type is None:
                        continue
                    one_day_jump_i = self.one_day_jump[fund_id]
                    three_day_jump_i = self.three_day_jump[fund_id]
                    if select_type in ['stock','QDII']:
                        if (one_day_jump_i < 0.1) & (three_day_jump_i < 0.25):
                            normal_fund_list.append(fund_id)
                        else:
                            self.abnormal_reason.append({'fund_id':fund_id,'abnormal_reason':f'high_risk: jump_max_1d {round(one_day_jump_i,3)} jump_max_3d {round(three_day_jump_i, 3)}'})
                    
                    elif select_type in ['bond','mix','mmf']:
                        if (one_day_jump_i < 0.05) & (three_day_jump_i < 0.15):
                            normal_fund_list.append(fund_id)
                        else:
                            self.abnormal_reason.append({'fund_id':fund_id,'abnormal_reason':f'low_risk: jump_max_1d {round(one_day_jump_i,3)} jump_max_3d {round(three_day_jump_i, 3)}'})
                else:
                    self.abnormal_reason.append({'fund_id':fund_id,'abnormal_reason':'price_inactive'})

            else:
                self.abnormal_reason.append({'fund_id':fund_id,'abnormal_reason':'less_than_half_y'})
        # 全部计算 
        self._fund_nav = self._fund_nav_origin.copy()#[normal_fund_list].copy()
        _tm_filter_nav = time.time()
        index_price: pd.DataFrame = self._basic_api.get_index_price_dt(start_date, end_date, columns=['close'])
        _tm_load_index = time.time()
        # 有的index最近没有数据，如活期存款利率/各种定期存款利率等，需要先reindex到全部的交易日列表上ffill
        self._index_price: pd.DataFrame = index_price.pivot_table(index='datetime', columns='index_id', values='close').reindex(self._trading_days.index).fillna(method='ffill')
        _bank_rate_df: pd.DataFrame = self._basic_api.get_index_price(index_list=['ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'])
        _bank_rate_df = _bank_rate_df.pivot_table(index='datetime', columns='index_id', values='close').reindex(all_trading_days.index).fillna(method='ffill').reindex(self._trading_days.index)
        drop_list = [i for i in self._index_price if i in _bank_rate_df]
        self._index_price = self._index_price.drop(drop_list, axis=1)
        self._index_price: pd.DataFrame = pd.concat([self._index_price, _bank_rate_df], axis=1)
        
        self.fund_type_benchmark = self._index_price[[self._REPLACE_DICT[fund_type] for fund_type in self._FUND_CLASSIFIER]]
        self.fund_type_benchmark = self.fund_type_benchmark.rename(columns={v : f'{k}_bench' for k, v in self._REPLACE_DICT.items()})
        # 再reindex到参与计算的交易日列表上
        # self._index_price: pd.DataFrame = index_price.reindex(self._trading_days.index)
        pd.testing.assert_index_equal(self._fund_nav.index, self._index_price.index)
        
        macroeco = RawDataApi().get_em_macroeconomic_daily(codes=['EX_DR_RATIO']).drop(columns='_update_time')
        self._macroeco: pd.DataFrame = macroeco.pivot_table(index='datetime', columns='codes', values='value').reindex(all_trading_days.index).fillna(method='ffill').reindex(self._trading_days.index)
        pd.testing.assert_index_equal(self._fund_nav.index, self._macroeco.index)
        try:
            # 这个指数有一天的点数是0，特殊处理一下
            self._index_price['spi_spa'] = self._index_price['spi_spa'].where(self._index_price.spi_spa != 0).fillna(method='ffill')
        except KeyError:
            pass

        # 对数净值取差分并且去掉第一个空值，得到基金对数收益率数列
        self._fund_ret: pd.DataFrame = np.log(self._fund_nav).diff().iloc[1:, :]

        self._fund_size: pd.DataFrame = self._basic_api.get_fund_size_range(start_date, end_date)
        self._fund_size = self._fund_size.pivot_table(index='datetime', columns='fund_id', values='size')
        self._fund_size = self._fund_size.reindex(self._fund_ret.index.union(self._fund_size.index)).fillna(method='ffill').reindex(self._fund_ret.index)
        pd.testing.assert_index_equal(self._fund_ret.index, self._fund_size.index)

        # 计算benchmark return，且index对齐到fund_ret
        self._fund_benchmark_df: pd.DataFrame = self._basic_api.get_fund_benchmark().drop(columns='_update_time')
        _tm_other_processor = time.time()
        benchmark_ret: pd.DataFrame = self._get_benchmark_return()
        _tm_other_processor_tmp = time.time()
        self._benchmark_ret: pd.DataFrame = benchmark_ret.reindex(self._fund_ret.index)
        benchmark_cols = self._benchmark_ret.columns.tolist()
        for fund_id in self._fund_nav.columns:
            benchmark_id = fund_id
            if not benchmark_id in benchmark_cols:    
                wind_class_2 = self._wind_class_2_dict[fund_id]
                select_type = self._wind_type_dict[wind_class_2]
                _benchmark_ret = self._index_price[self._REPLACE_DICT[select_type]].pct_change(1).rename(benchmark_id)
                self._benchmark_ret = self._benchmark_ret.join(_benchmark_ret)

        pd.testing.assert_index_equal(self._fund_ret.index, self._benchmark_ret.index)
        self._fund_ret = self._fund_ret.join(self._benchmark_ret, rsuffix='_b')
        _tm_benchmakr_ret = time.time()
        # 获取wind一级分类
        self._wind_class_1: np.ndarray = self._fund_list.wind_class_1.unique()
        # 按wind二级分类计算基金收益率排名
        self._get_continue_stats_tool()
        # 取出沪深300收益率
        self._hs300 = self._basic_api.get_index_price(['hs300'])[['datetime', 'close']]
        self._hs300 = self._hs300.set_index('datetime')['close']
        self._hs300 = self._hs300.apply(np.log).diff()[1:]
        self._fund_ret = self._fund_ret.join(self._hs300.rename('close_hs300'))
        # 取出货币基金指数收益率
        self._mmf = self._basic_api.get_index_price(['mmf'])[['datetime', 'close']]
        self._mmf = self._mmf.set_index('datetime')['close']
        self._mmf = self._mmf.apply(np.log).diff()[1:]
        self._fund_ret = self._fund_ret.join(self._mmf.rename('close_mmf'))
        self._fund_ret = self._fund_ret.join(self.fund_type_benchmark.pct_change(1))
        # 获取年报/半年报的发布日期
        end_date_dt = pd.to_datetime(end_date, infer_datetime_format=True)
        md = end_date_dt.strftime('%m%d')
        sentry_date = pd.Series(self._REPORT_DATE_LIST)
        sentry = sentry_date[sentry_date <= md]
        # 为半年报/年报发布预留一个季度的缓冲时间
        if sentry.empty:
            real_date = datetime.date(end_date_dt.year - 1, 6, 30)
        elif len(sentry) <= 2:
            real_date = datetime.date(end_date_dt.year - 1, 12, 31)
        else:
            real_date = datetime.date(end_date_dt.year, 6, 30)

        # 取出持仓数据
        self._fund_hold = self._basic_api.get_history_fund_size(start_date=real_date, end_date=real_date)
        self._fund_hold = self._fund_hold[['fund_id', 'institution_holds', 'hold_num']].set_index('fund_id')
        # 取出货基风险相关数据
        self._risk_metric = self._basic_api.get_fund_hold_asset_by_id(start_date=real_date, end_date=real_date)
        self._risk_metric = self._risk_metric[['fund_id', 'first_repo_to_nav', 'avg_ptm']].set_index('fund_id')
        _tm_ret_process = time.time()
        if is_print_time:
            print(f'\t[time] fetch fund nav : {_tm_loads_nav_1 - _tm_loads_nav_0}')
            print(f'\t[time] filter nav : {_tm_filter_nav - _tm_loads_nav_1}')
            print(f'\t[time] fetch index price : {_tm_load_index - _tm_filter_nav}')
            print(f'\t[time] fetch fund size and benchmark info : {_tm_other_processor - _tm_load_index}')
            print(f'\t[time] calc fund benchmark ret : {_tm_other_processor_tmp - _tm_other_processor}')
            print(f'\t[time] add fund benchmark ret : {_tm_benchmakr_ret - _tm_other_processor_tmp}')
            print(f'\t[time] combine fund rets : {_tm_ret_process - _tm_benchmakr_ret}')
            
    def history_load_data(self, end_date):
        self._wind_type_dict = {}
        for type_i, type_list in self._FUND_CLASSIFIER.items():
            for _ in type_list:
                self._wind_type_dict.update({_:type_i})
        # 获取待计算的基金列表
        self.fund_info: pd.DataFrame = self._basic_api.get_fund_info().drop(columns='_update_time')
        self._wind_class_2_dict = self.fund_info.set_index('fund_id')['wind_class_2'].to_dict()
        self._fund_list: pd.DataFrame = self.fund_info[self.fund_info.structure_type <= 1]
       
        # 获取区间内交易日列表
        self._trading_days_all: pd.Series = self._basic_api.get_trading_day_list().drop(columns='_update_time').set_index('datetime')
        # 基金净值数据、指数价格数据的index对齐到交易日列表
        self._fund_nav_all: pd.DataFrame = self._basic_api.get_fund_nav_with_date(end_date=end_date,fund_list=self._fund_list.fund_id.tolist())
        self._fund_nav_all = self._fund_nav_all.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value')
        self._index_price_all: pd.DataFrame = self._basic_api.get_index_price_dt(end_date=end_date, columns=['close'])
        self._index_price_all = self._index_price_all.pivot_table(index='datetime', columns='index_id', values='close')
        _bank_rate_df: pd.DataFrame = self._basic_api.get_index_price(index_list=['ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'])
        self._bank_rate_df = _bank_rate_df.pivot_table(index='datetime', columns='index_id', values='close').reindex(self._trading_days_all.index).fillna(method='ffill')
        drop_list = [i for i in self._index_price_all if i in self._bank_rate_df]
        self._index_price_all = self._index_price_all.drop(columns=drop_list)
        self._macroeco_all = RawDataApi().get_em_macroeconomic_daily(codes=['EX_DR_RATIO']).drop(columns='_update_time')
        self._macroeco_all = self._macroeco_all.pivot_table(index='datetime', columns='codes', values='value').reindex(self._trading_days_all.index).fillna(method='ffill')
        self._fund_size: pd.DataFrame = self._basic_api.get_fund_size_range(end_date=end_date)
        self._fund_size = self._fund_size.pivot_table(index='datetime', columns='fund_id', values='size')
        # 获取待计算的基金列表
        fund_list: pd.DataFrame = self._basic_api.get_fund_info().drop(columns='_update_time')
        self._fund_list: pd.DataFrame = fund_list[fund_list.structure_type <= 1]
        # 获取wind一级分类
        self._wind_class_1: np.ndarray = fund_list.wind_class_1.unique()
        # 取出沪深300收益率
        self._hs300 = self._basic_api.get_index_price(['hs300'])[['datetime', 'close']]
        self._hs300 = self._hs300.set_index('datetime')['close']
        self._hs300 = self._hs300.apply(np.log).diff()[1:]
        # 取出货币基金指数收益率
        self._mmf = self._basic_api.get_index_price(['mmf'])[['datetime', 'close']]
        self._mmf = self._mmf.set_index('datetime')['close']
        self._mmf = self._mmf.apply(np.log).diff()[1:]
        self._fund_benchmark_df: pd.DataFrame = self._basic_api.get_fund_benchmark().drop(columns='_update_time')

    def history_init(self, end_date):
        start_date: datetime.date = (pd.to_datetime(end_date, infer_datetime_format=True) - DateOffset(years=self._data_cycle+1)).date()
        print(f'start date for data: {start_date}')
        end_date: datetime.date = pd.to_datetime(end_date, infer_datetime_format=True).date()

        _trading_days: pd.Series = self._trading_days_all.loc[start_date:end_date]
        # 使用实际trading day的最后一天作为end_date
        self.end_date = _trading_days.index.array[-1]
        # 基金净值数据、指数价格数据的index对齐到交易日列表
        fund_nav: pd.DataFrame = self._fund_nav_all.loc[start_date:end_date, :]
        self._fund_nav_origin: pd.DataFrame = fund_nav.reindex(_trading_days.index).fillna(method='ffill')
        self.one_day_jump = self._fund_nav_origin.pct_change(1).abs().max()
        self.three_day_jump = self._fund_nav_origin.pct_change(3).abs().max()
        normal_fund_list = []
        self.abnormal_reason = []
        for fund_id in self._fund_nav_origin:
            _df = self._fund_nav_origin[fund_id].dropna()
            if _df.shape[0] > 120:
                if _df.iloc[-120:].var() > 1e-6:
                    wind_class_2 = self._wind_class_2_dict[fund_id]
                    select_type = self._wind_type_dict.get(wind_class_2, None)
                    if select_type is None:
                        continue
                    one_day_jump_i = self.one_day_jump[fund_id]
                    three_day_jump_i = self.three_day_jump[fund_id]
                    if select_type in ['stock','QDII']:
                        if (one_day_jump_i < 0.1) & (three_day_jump_i < 0.25):
                            normal_fund_list.append(fund_id)
                        else:
                            self.abnormal_reason.append({'fund_id':fund_id,'abnormal_reason':f'high_risk: jump_max_1d {round(one_day_jump_i,3)} jump_max_3d {round(three_day_jump_i, 3)}'})
                    
                    elif select_type in ['bond','mix','mmf']:
                        if (one_day_jump_i < 0.05) & (three_day_jump_i < 0.15):
                            normal_fund_list.append(fund_id)
                        else:
                            self.abnormal_reason.append({'fund_id':fund_id,'abnormal_reason':f'low_risk: jump_max_1d {round(one_day_jump_i,3)} jump_max_3d {round(three_day_jump_i, 3)}'})
                else:
                    self.abnormal_reason.append({'fund_id':fund_id,'abnormal_reason':'price_inactive'})

            else:
                self.abnormal_reason.append({'fund_id':fund_id,'abnormal_reason':'less_than_half_y'})
        self._fund_nav = self._fund_nav_origin[normal_fund_list].copy()
        index_price: pd.DataFrame = self._index_price_all.loc[start_date:end_date, :]
        # 有的index最近没有数据，如活期存款利率/各种定期存款利率等，需要先reindex到全部的交易日列表上ffill
        self._index_price: pd.DataFrame = index_price.reindex(_trading_days.index).fillna(method='ffill')
        _bank_rate_df = self._bank_rate_df.reindex(_trading_days.index)
        self._index_price: pd.DataFrame = pd.concat([self._index_price, _bank_rate_df], axis=1).dropna(how='all', axis=1)
        # 再reindex到参与计算的交易日列表上
        # self._index_price: pd.DataFrame = index_price.reindex(_trading_days.index)
        pd.testing.assert_index_equal(self._fund_nav.index, self._index_price.index)
        
        self._macroeco: pd.DataFrame = self._macroeco_all.reindex(_trading_days.index)
        pd.testing.assert_index_equal(self._fund_nav.index, self._macroeco.index)
        try:
            # 这个指数有一天的点数是0，特殊处理一下
            self._index_price['spi_spa'] = self._index_price['spi_spa'].where(self._index_price.spi_spa != 0).fillna(method='ffill')
        except KeyError:
            pass

        # 对数净值取差分并且去掉第一个空值，得到基金对数收益率数列
        self._fund_ret: pd.DataFrame = np.log(self._fund_nav).diff().iloc[1:, :]

        self._fund_size = self._fund_size.reindex(self._fund_ret.index.union(self._fund_size.index)).fillna(method='ffill').reindex(self._fund_ret.index)
        pd.testing.assert_index_equal(self._fund_ret.index, self._fund_size.index)

        # 计算benchmark return，且index对齐到fund_ret
        self._benchmark_ret: pd.DataFrame = self._get_benchmark_return()
        self._benchmark_ret = self._benchmark_ret.reindex(self._fund_ret.index).dropna(how='all', axis=1)
        benchmark_cols = self._benchmark_ret.columns.tolist()
        for fund_id in self._fund_nav.columns:
            benchmark_id = fund_id
            if not benchmark_id in benchmark_cols:    
                wind_class_2 = self._wind_class_2_dict[fund_id]
                select_type = self._wind_type_dict[wind_class_2]
                _benchmark_ret = self._index_price[self._REPLACE_DICT[select_type]].pct_change(1).rename(benchmark_id)
                self._benchmark_ret = self._benchmark_ret.join(_benchmark_ret)
        pd.testing.assert_index_equal(self._fund_ret.index, self._benchmark_ret.index)
        self._fund_ret = self._fund_ret.join(self._benchmark_ret, rsuffix='_b')
        # 按wind二级分类计算基金收益率排名
        self._get_continue_stats_tool()

        self._fund_ret = self._fund_ret.join(self._hs300.rename('close_hs300'))
        self._fund_ret = self._fund_ret.join(self._mmf.rename('close_mmf'))
        self._fund_ret = self._fund_ret.dropna(how='all', axis=1)
        
        self.fund_type_benchmark = self._index_price[[self._REPLACE_DICT[fund_type] for fund_type in self._FUND_CLASSIFIER]]
        self.fund_type_benchmark = self.fund_type_benchmark.rename(columns={v : f'{k}_bench' for k, v in self._REPLACE_DICT.items()})
        self._fund_ret = self._fund_ret.join(self.fund_type_benchmark.pct_change(1))
        # 获取年报/半年报的发布日期
        md = end_date.strftime('%m%d')
        sentry_date = pd.Series(self._REPORT_DATE_LIST)
        sentry = sentry_date[sentry_date <= md]
        # 为半年报/年报发布预留一个季度的缓冲时间
        if sentry.empty:
            real_date = datetime.date(end_date.year - 1, 6, 30)
        elif len(sentry) <= 2:
            real_date = datetime.date(end_date.year - 1, 12, 31)
        else:
            real_date = datetime.date(end_date.year, 6, 30)

        # 取出持仓数据
        self._fund_hold = self._basic_api.get_history_fund_size(start_date=real_date, end_date=real_date)
        self._fund_hold = self._fund_hold[['fund_id', 'institution_holds', 'hold_num']].set_index('fund_id')
        # 取出货基风险相关数据
        self._risk_metric = self._basic_api.get_fund_hold_asset_by_id(start_date=real_date, end_date=real_date)
        self._risk_metric = self._risk_metric[['fund_id', 'first_repo_to_nav', 'avg_ptm']].set_index('fund_id')

    def df_resample(self, df):
        df.index = pd.to_datetime(df.index)
        df = df.resample('1W').sum()
        df.index = df.index.date
        return df

    def stutzer_index(self, ex_return: np.ndarray):
        information_statistic = lambda theta: np.log(np.mean(np.exp(theta[0] * ex_return)))
        theta0 = [-1.]
        bounds = Bounds((-np.inf), (0))
        result = minimize(information_statistic, theta0, method='SLSQP', bounds=bounds, tol=1e-16)
        if result.success:
            information_statistic = -result.fun
            if information_statistic <= 0:
                return 0
            stutzer_index = np.abs(np.mean(ex_return)) / np.mean(ex_return) * np.sqrt(2 * information_statistic)
            # result = {'information_statistic': information_statistic,
            #         'stutzer_index': stutzer_index,
            #         'theta': result.x[0]}
            return stutzer_index
        else:
            # result = {'information_statistic': information_statistic,
            #
            #   'stutzer_index': np.nan,
            #         'theta': result.x[0]}
            # print('未找到Information Statistic最大值')
            return np.nan

    def _lambda_cl(self, total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X[:, 0][X[:, 0] < 0] = 0
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                'alpha_t':est2.tvalues[-1],
                'alpha_p':est2.pvalues[-1],
                # TODO beta 2 - beta 1 , p values and t values calculation is not correct
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }

    def _lambda_tm(self, total: np.ndarray):
        if not total.size:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1] * total[:, 1]]).T
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[1],
                'alpha':est2.params[-1],
                'alpha_t':est2.tvalues[-1],
                'alpha_p':est2.pvalues[-1],
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }

    def _lambda_hm(self, total: np.ndarray):
        if not total.size:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    'alpha_t':np.nan,
                    'alpha_p':np.nan,
                    'beta_t':np.nan,
                    'beta_p':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 1][X[:, 1] < 0] = 0
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[1],
                'alpha':est2.params[-1],
                'alpha_t':est2.tvalues[-1],
                'alpha_p':est2.pvalues[-1],
                'beta_t':est2.tvalues[1],
                'beta_p':est2.pvalues[1]
                }

    def _lambda_alpha_beta(self, df_i):
        if np.count_nonzero(np.count_nonzero(df_i, axis=0)) < df_i.shape[1]:
            return np.nan, 0
        ploy_res = np.polyfit(y=df_i[:, 0], x=df_i[:, 1], deg=1)
        beta = ploy_res[0]
        alpha = ploy_res[1] * self._TRADING_DAYS_PER_YEAR
        return alpha, beta

    def _get_benchmark_return(self) -> pd.DataFrame:
        benchmark_list: Dict[str, float] = {}
        # 遍历每一只基金的benchmark进行处理
        for row in self._fund_benchmark_df.itertuples(index=False):
            values: List[pd.Series] = []
            cons: float = 0
            # 空的benchmark表明我们没有对应的指数或无法解析公式
            if row.benchmark_s:
                benchmark: Dict[str, float] = json.loads(row.benchmark_s)
                benchmark_raw: Dict[str, float] = eval(row.benchmark)
                for (index, weight), index_raw in zip(benchmark.items(), benchmark_raw.keys()):
                    if index == '1':
                        # 表示benchmark中该项为常数
                        cons += weight
                    elif index in ('ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'):
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            # self.log.append((row.fund_id, index))
                            break
                        else:
                            try:
                                if index == 'nonor':
                                    # 在这里我们用超额存款准备金率替代同业存款利率
                                    ra: pd.Series = self._macroeco.loc[:, 'EX_DR_RATIO']
                                else:
                                    ra: pd.Series = self._index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'[benchmark_return] Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                # self.log.append((row.fund_id, index))
                                break
                            else:
                                values.append(ra.iloc[1:] * 0.01 * weight / self._TRADING_DAYS_PER_YEAR)
                    else:
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            # self.log.append((row.fund_id, index))
                            break
                        else:
                            try:
                                ra: pd.Series = self._index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                # self.log.append((row.fund_id, index))
                                break
                            else:
                                ra = np.log(ra).diff().iloc[1:]
                                values.append(ra * weight)
                else:
                    if values or cons:
                        the_sum: float = sum(values)
                        if cons:
                            the_sum += np.log(math.pow(1 + cons, 1 / self._TRADING_DAYS_PER_YEAR))
                        benchmark_list[row.fund_id] = the_sum

        return pd.DataFrame.from_dict(benchmark_list)

    def _lambda_make_rank_dic(self, x: pd.DataFrame):
        resample_ret: pd.DataFrame = self.df_resample(self._fund_ret.loc[:, self._fund_ret.columns.intersection(x.fund_id.array)])
        self._rank_dic[x.name] = resample_ret.rank(pct=True, axis=1).rename_axis(columns='fund_id')

    def _get_continue_stats_tool(self):
        self._fund_list.groupby(by='wind_class_2').apply(self._lambda_make_rank_dic)

    def get_scale(self, end_date, fund_id):
        try:
            scale = self._fund_size.loc[self.end_date, fund_id]
        except KeyError:
            scale = np.nan
        return scale

    def _calculate_item(self, fund_id: str, wind_2_class: str):
        # 取最后一年的有效净值
        _sr: pd.Series = self._fund_nav[fund_id]
        sr: pd.Series = _sr[_sr.notna()].tail(self._data_cycle_days)
        temp_nav = _sr.to_numpy()
        # 生成一个m*n的矩阵
        temp_nav = np.array([temp_nav[i:self._TRADING_DAYS_PER_YEAR//2 + i + 20] for i in range(0, temp_nav.shape[0] - 20 - self._TRADING_DAYS_PER_YEAR//2, 20)])
        ret: np.ndarray = temp_nav[:, 1:] / temp_nav[:, :-1]
        # 过滤掉全是nan或只有1个不是nan的行
        filter1 = (ret.shape[1] - np.count_nonzero(np.isnan(ret), axis=1) > 1)
        ret, temp_nav = ret[filter1], temp_nav[filter1]
        annual_vol: np.ndarray = np.nanstd(ret, axis=1, ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        # 一维数组，所以这里直接用nonzero即可
        filter1 = annual_vol.nonzero()
        annual_vol, temp_nav = annual_vol[filter1], temp_nav[filter1]
        annual_ret = np.exp(np.log(temp_nav[:, -1] / temp_nav[:, 0]) / temp_nav.shape[1] * self._TRADING_DAYS_PER_YEAR) - 1
        _res = (annual_ret - self._RISK_FEE_RATE) / annual_vol
        _res = _res[~np.isnan(_res)]
        if _res.size < 6:
            continue_regress_v = np.nan
            continue_regress_t = np.nan
        else:
            mod = AutoReg(_res, 1)
            res = mod.fit()
            continue_regress_v = res.params[0]
            continue_regress_t = res.tvalues[0]

        if wind_2_class is not None:
            _rank_df = self._rank_dic[wind_2_class]
            res = []
            last = ''
            for idx, i in enumerate(_rank_df[fund_id]):
                now = 'W' if i > 0.5 else 'L'
                if idx != 0:
                    res.append(last+now)
                last = now
            ww = max(res.count('WW'), 1)
            ll = max(res.count('LL'), 1)
            wl = max(res.count('WL'), 1)
            lw = max(res.count('LW'), 1)
            continue_stats_v = ww * ll / wl / lw
        else:
            continue_stats_v = np.nan

        # 取最后一年的有效收益率+业绩比较基准收益率
        df_i: np.ndarray = self._fund_ret[[fund_id, fund_id+'_b']].to_numpy()
        df_i = df_i[(~np.isnan(df_i)).all(axis=1)]
        df_i = df_i[-self._data_cycle_days:]

        df_i_hs300: np.ndarray = self._fund_ret[[fund_id, 'close_hs300']].to_numpy()
        df_i_hs300 = df_i_hs300[(~np.isnan(df_i_hs300)).all(axis=1)]
        df_i_hs300 = df_i_hs300[-self._data_cycle_days:]

        df_i_mmf: np.ndarray = self._fund_ret[[fund_id, 'close_mmf']].to_numpy()
        df_i_mmf = df_i_mmf[(~np.isnan(df_i_mmf)).all(axis=1)]
        df_i_mmf = df_i_mmf[-self._data_cycle_days:]

        wind_class_2 = self._wind_class_2_dict[fund_id]
        fund_type = self._wind_type_dict[wind_class_2]
        df_i_cl = self._fund_ret[[fund_id]].join(self._fund_ret[[f'{fund_type}_bench']]).to_numpy()
        df_i_cl = df_i_cl[(~np.isnan(df_i_cl)).all(axis=1)]
        df_i_cl = df_i_cl[-self._data_cycle_days:]

        # alpha, beta
        alpha, beta = self._lambda_alpha_beta(df_i)
        # alpha, beta to the market
        alpha_hs300, beta_hs300 = self._lambda_alpha_beta(df_i_hs300)
        # mdd & len(mdd)
        mdd_part = sr / sr.cummax()
        mdd = 1 - mdd_part.min()
        mdd_date2 = mdd_part.idxmin()
        mdd_date1 = sr[:mdd_date2].idxmax()
        mdd_len = (mdd_date2 - mdd_date1).days
        # sharpe ratio
        annual_vol = (sr / sr.shift(1)).std(ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        annual_ret = np.exp(np.log(sr.iloc[-1] / sr.fillna(method='bfill').iloc[0]) / sr.shape[0] * self._TRADING_DAYS_PER_YEAR) - 1
        sharpe = (annual_ret - self._RISK_FEE_RATE) / annual_vol
        # treynor ratio
        if beta != 0:
            treynor = (np.nanmean(df_i[:, 0]) - self._RISK_FEE_RATE_PER_DAY) / beta
        else:
            treynor = np.nan
        # calma ratio
        calma_ratio = annual_ret / mdd if mdd != 0 else np.nan
        # ex return
        _ex_return = df_i[:, 0] - df_i[:, 1]
        # excess return against hs300
        _ex_return_hs300 = df_i_hs300[:, 0] - df_i_hs300[:, 1]
        # excess return against mmf
        _ex_return_mmf = df_i_mmf[:, 0] - df_i_mmf[:, 1]
        # winning rate against mmf
        if len(_ex_return_mmf) != 0:
            winning_rate = len(_ex_return_mmf[_ex_return_mmf > 0]) / len(_ex_return_mmf)
        else:
            winning_rate = 0
        # track error
        track_err = np.nanstd(_ex_return, ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        # track error agains hs300
        track_err_hs300 = np.nanstd(_ex_return_hs300, ddof=1) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        # information error
        info_ratio = alpha / track_err
        # information ratio against hs300
        info_ratio_hs300 = alpha_hs300 / track_err_hs300
        # scale
        scale = self.get_scale(self.end_date, fund_id)
        # VaR
        ret = df_i[:, 0]
        if ret.size > 0:
            VaR = np.quantile(ret, 0.05)
            CVaR = ret[ret <= VaR].mean()
        else:
            VaR = np.nan
            CVaR = np.nan
        # institution holds & hold_num
        try:
            temp_hold = self._fund_hold.loc[fund_id, :]
            ins_holds = temp_hold.institution_holds
            hold_num = temp_hold.hold_num
        except KeyError:
            ins_holds = np.nan
            hold_num = np.nan
        #
        try:
            temp_risk = self._risk_metric.loc[fund_id, :]
            leverage = temp_risk.first_repo_to_nav
            ptm = temp_risk.avg_ptm
        except KeyError:
            leverage = np.nan
            ptm = np.nan

        stock_tm_res = self._lambda_tm(df_i_cl)
        stock_tm_alpha = stock_tm_res['alpha']
        stock_tm_alpha_t = stock_tm_res['alpha_t']
        stock_tm_alpha_p = stock_tm_res['alpha_p']
        stock_tm_beta = stock_tm_res['beta']
        stock_tm_beta_t = stock_tm_res['beta_t']
        stock_tm_beta_p = stock_tm_res['beta_p']

        stock_hm_res = self._lambda_hm(df_i_cl)
        stock_hm_alpha = stock_hm_res['alpha']
        stock_hm_alpha_t = stock_hm_res['alpha_t']
        stock_hm_alpha_p = stock_hm_res['alpha_p']
        stock_hm_beta = stock_hm_res['beta']
        stock_hm_beta_t = stock_hm_res['beta_t']
        stock_hm_beta_p = stock_hm_res['beta_p']

        stock_cl_res = self._lambda_cl(df_i_cl)
        stock_cl_alpha = stock_cl_res['alpha']
        stock_cl_alpha_t = stock_cl_res['alpha_t']
        stock_cl_alpha_p = stock_cl_res['alpha_p']
        stock_cl_beta = stock_cl_res['beta']
        stock_cl_beta_t = stock_cl_res['beta_t']
        stock_cl_beta_p = stock_cl_res['beta_p']

        # stutzer
        stutzer = self.stutzer_index(df_i[:, 0]-self._RISK_FEE_RATE_PER_DAY) * np.sqrt(self._TRADING_DAYS_PER_YEAR)
        # stutzer = self.stutzer_index(np.array(self._fund_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id] - self._benchmark_ret.tail(self._TRADING_DAYS_PER_YEAR)[fund_id]))
        return pd.Series({
            'fund_id':fund_id,
            'datetime':self.end_date,
            'alpha':alpha,
            'beta':beta,
            'annual_vol':annual_vol,
            'annual_ret':annual_ret,
            'track_err':track_err,
            'mdd': mdd,
            'mdd_len':mdd_len,
            'winning_rate':winning_rate,
            'VaR':VaR,
            'CVaR':CVaR,
            'ins_holds':ins_holds,
            'hold_num':hold_num,
            'leverage':leverage,
            'ptm':ptm,
            'continue_regress_v':continue_regress_v,
            'continue_regress_t':continue_regress_t,
            'continue_stats_v':continue_stats_v,

            'stock_cl_alpha':stock_cl_alpha,
            'stock_cl_alpha_t':stock_cl_alpha_t,
            'stock_cl_alpha_p':stock_cl_alpha_p,
            'stock_cl_beta':stock_cl_beta,
            'stock_cl_beta_t':stock_cl_beta_t,
            'stock_cl_beta_p':stock_cl_beta_p,

            'stock_tm_alpha':stock_tm_alpha,
            'stock_tm_alpha_t':stock_tm_alpha_t,
            'stock_tm_alpha_p':stock_tm_alpha_p,
            'stock_tm_beta':stock_tm_beta,
            'stock_tm_beta_t':stock_tm_beta_t,
            'stock_tm_beta_p':stock_tm_beta_p,

            'stock_hm_alpha':stock_hm_alpha,
            'stock_hm_alpha_t':stock_hm_alpha_t,
            'stock_hm_alpha_p':stock_hm_alpha_p,
            'stock_hm_beta':stock_hm_beta,
            'stock_hm_beta_t':stock_hm_beta_t,
            'stock_hm_beta_p':stock_hm_beta_p,
            'info_ratio':info_ratio,
            'info_ratio_hs300':info_ratio_hs300,
            'treynor':treynor,
            'sharpe':sharpe,
            'calma_ratio':calma_ratio,
            'stutzer':stutzer,
            'scale':scale,
        })

    def calculate(self):
        fund_list = self._fund_list[['fund_id', 'wind_class_2']]
        # 对fund_nav, fund_list二者交集的每个fund_id进行计算
        self._result = fund_list[fund_list.fund_id.isin(self._fund_nav.columns.to_list())].apply(lambda x: self._calculate_item(x.fund_id, x.wind_class_2), axis=1)
        self._result['ins_holds'] = self._result['ins_holds'].fillna(0)
        self._result['hold_num'] = self._result['hold_num'].fillna(0)
        # add abnormal
        _df = pd.DataFrame(self.abnormal_reason)
        _df['datetime'] = self.end_date
        self._result = pd.merge(left=self._result, right=_df, how='outer', on=['fund_id','datetime'])
        self._result['data_cycle'] = str(self._data_cycle) + 'Y'
        self._result = self._result.replace({np.inf: np.nan, -np.inf: np.nan})

    def process(self, start_date: str, end_date: str, end_date_dt: datetime.date) -> List[str]:
        print(f'indicator group update on the last day of week {end_date_dt}')
        failed_tasks: List[str] = []
        try:
            self.init(end_date)
            print('init done, to calc indicators')
            self.calculate()
            self._data_helper._upload_derived(self._result, FundIndicatorGroup.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'fund_indicator_group({self._data_cycle}Y)')
        return failed_tasks

    @staticmethod
    def update_history(end_date: str, data_cycle: int):
        from ...wrapper.mysql import BasicDatabaseConnector
        from ...view.basic_models import TradingDayList
        with BasicDatabaseConnector().managed_session() as quant_session:
            query = quant_session.query(TradingDayList)
            trade_days = pd.read_sql(query.statement, query.session.bind)

        fipg = FundIndicatorProcessorGroup(DerivedDataHelper(), data_cycle)
        fipg.history_load_data(end_date)
        dts = trade_days[(trade_days.datetime >= datetime.date(2010, 1, 1)) &
                         (trade_days.datetime < pd.to_datetime(end_date, infer_datetime_format=True))].datetime.tolist()
        dts.reverse()
        dts = [dt for dt in dts if dt.weekday() == 4]
        for date in dts:
            fipg.history_init(date)
            fipg.calculate()
            fipg._data_helper._upload_derived(fipg._result, FundIndicatorGroup.__table__.name)
            print(f'{date} done')


if __name__ == '__main__':
    date = '20210129'
    fipg = FundIndicatorProcessorGroup(DerivedDataHelper(), 5)
    fipg.process(date, date, pd.to_datetime(date).date())
    # FundIndicatorProcessorGroup.update_history('20201009', 5)
